""" This submodule implements some useful classes for manipulating
DOS and dispersion data.

"""

from .dos import VibrationalDOS, ElectronicDOS, DensityOfStates
from .dispersion import VibrationalDispersion, ElectronicDispersion, Dispersion
from .spectral import Spectral


__all__ = [
    'VibrationalDOS',
    'ElectronicDOS',
    'DensityOfStates',
    'VibrationalDispersion',
    'ElectronicDispersion',
    'Dispersion',
    'Spectral'
]



#!/usr/bin/env python
import unittest
import re

from matador.query import DBQuery
from matador.utils.chem_utils import parse_element_string


class QueryParseTest(unittest.TestCase):
    """ Test query functionality. """

    def test_basic_queries(self):
        kwargs = {"composition": "KP", "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["K"]}},
                        {"elems": {"$in": ["P"]}},
                        {"stoichiometry": {"$size": 2}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {"formula": "K3P4", "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"stoichiometry": {"$in": [["K", 3]]}},
                        {"stoichiometry": {"$in": [["P", 4]]}},
                        {"stoichiometry": {"$size": 2}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {"formula": "K3P6", "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"stoichiometry": {"$in": [["K", 1]]}},
                        {"stoichiometry": {"$in": [["P", 2]]}},
                        {"stoichiometry": {"$size": 2}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {"formula": "K27P9", "partial_formula": True, "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"stoichiometry": {"$in": [["K", 3]]}},
                        {"stoichiometry": {"$in": [["P", 1]]}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {"formula": ["K27P9"], "partial_formula": True, "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"stoichiometry": {"$in": [["K", 3]]}},
                        {"stoichiometry": {"$in": [["P", 1]]}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {
            "field": ["xc_functional", "cut_off_energy"],
            "filter": [["PBE"], ["301.2312", 400.0]],
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {"xc_functional": "PBE"},
                {"cut_off_energy": {"$gte": 301.2312, "$lte": 400.0}},
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {
            "field": ["cut_off_energy", "xc_functional"],
            "filter": [["301.2312", 400.0], ["PBE", "LDA"]],
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {"cut_off_energy": {"$gte": 301.2312, "$lte": 400.0}},
                {"$or": [{"xc_functional": "PBE"}, {"xc_functional": "LDA"}]},
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_complex_queries(self):
        """ Test long queries with multiple mismatching of lists
        (emulating argparse) and values.
        """
        self.maxDiff = None
        kwargs = {
            "formula": "K27P9",
            "space_group": ["Fd-3m"],
            "num_species": 3,
            "num_fu": [4],
            "tags": ["foo", "Bar"],
            "doi": ["1001/4001"],
            "icsd": 100020,
            "cutoff": 300,
            "spin": 1,
            "partial_formula": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"stoichiometry": {"$in": [["K", 3]]}},
                        {"stoichiometry": {"$in": [["P", 1]]}},
                    ]
                },
                {"stoichiometry": {"$size": 3}},
                {"space_group": "Fd-3m"},
                {"num_fu": {"$gte": 4}},
                {"$and": [{"tags": {"$in": ["foo"]}}, {"tags": {"$in": ["Bar"]}}]},
                {"doi": {"$in": ["1001/4001"]}},
                {"$or": [{'icsd': {"$eq": "100020"}}, {'icsd': {'$eq': 100020}}]},
                {"cut_off_energy": {"$eq": 300}},
                {"spin_polarized": True},
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {
            "composition": "LiFeBe",
            "icsd": True,
            "ignore_warnings": True,
            "src_str": "/Foo/bar/foo/Bar.res",
            "pressure": 5,
            "pressure_tolerance": 1,
            "cutoff": [300, 400],
            "encapsulated": True,
            "cnt_radius": 5.21,
            "sedc": "null",
            "mp_spacing": [0.05],
            "spin": 0,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["Li"]}},
                        {"elems": {"$in": ["Fe"]}},
                        {"elems": {"$in": ["Be"]}},
                        {"stoichiometry": {"$size": 3}},
                    ]
                },
                {"icsd": {"$exists": True}},
                {"cut_off_energy": {"$gte": 300, "$lte": 400}},
                {"source": {"$in": [re.compile("/Foo/bar/foo/Bar.res")]}},
                {"pressure": {"$lte": 6.0, "$gte": 4.0}},
                {"encapsulated": {"$exists": True}},
                {"cnt_radius": {"$gte": 5.20, "$lte": 5.22}},
                {"sedc_scheme": {"$exists": False}},
                {"kpoints_mp_spacing": {"$gte": 0.04, "$lte": 0.06}},
                {"spin_polarized": {"$ne": True}},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {
            "composition": "LiFeBe",
            "icsd": False,
            "ignore_warnings": True,
            "src_str": "/Foo/bar/foo/Bar.res",
            "pressure": 5,
            "pressure_tolerance": None,
            "kpoint_tolerance": None,
            "cutoff": [300, 400],
            "encapsulated": True,
            "cnt_radius": 5.21,
            "sedc": "null",
            "mp_spacing": [0.05],
            "spin": "any",
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["Li"]}},
                        {"elems": {"$in": ["Fe"]}},
                        {"elems": {"$in": ["Be"]}},
                        {"stoichiometry": {"$size": 3}},
                    ]
                },
                {"icsd": {"$exists": False}},
                {"cut_off_energy": {"$gte": 300, "$lte": 400}},
                {"source": {"$in": [re.compile("/Foo/bar/foo/Bar.res")]}},
                {"pressure": {"$lte": 5.5, "$gte": 4.5}},
                {"encapsulated": {"$exists": True}},
                {"cnt_radius": {"$gte": 5.20, "$lte": 5.22}},
                {"sedc_scheme": {"$exists": False}},
                {"kpoints_mp_spacing": {"$lte": 0.06, "$gte": 0.04}},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {
            "composition": "LiFeBe",
            "icsd": True,
            "ignore_warnings": True,
            "src_str": "/Foo/bar/foo/Bar.res",
            "pressure": 5,
            "pressure_tolerance": 10,
            "kpoint_tolerance": 0.02,
            "cutoff": [300, 400],
            "encapsulated": True,
            "cnt_radius": 5.21,
            "sedc": "null",
            "mp_spacing": [0.04, 0.05],
            "spin": "any",
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["Li"]}},
                        {"elems": {"$in": ["Fe"]}},
                        {"elems": {"$in": ["Be"]}},
                        {"stoichiometry": {"$size": 3}},
                    ]
                },
                {"icsd": {"$exists": True}},
                {"cut_off_energy": {"$gte": 300, "$lte": 400}},
                {"source": {"$in": [re.compile("/Foo/bar/foo/Bar.res")]}},
                {"pressure": {"$lte": 15, "$gte": -5}},
                {"encapsulated": {"$exists": True}},
                {"cnt_radius": {"$gte": 5.20, "$lte": 5.22}},
                {"sedc_scheme": {"$exists": False}},
                {"kpoints_mp_spacing": {"$lte": 0.05, "$gte": 0.04}},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_non_elemental_query(self):
        kwargs = {"composition": "Li4Ge:Fe2Be", "testing": True}
        query = DBQuery(**kwargs)

        kwargs = {"composition": "LiGeFeBe", "intersection": True, "testing": True}
        query2 = DBQuery(**kwargs)
        print(query.query_dict)
        print(query2.query_dict)

        self.assertDictEqual(query.query_dict, query2.query_dict)
        self.assertEqual(query._chempots, ["Li4Ge", "Fe2Be"])

    def test_duplicate_elements(self):
        kwargs = {"composition": "PPPPP", "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {"$and": [{"elems": {"$in": ["P"]}}, {"stoichiometry": {"$size": 1}}]},
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }

        self.assertDictEqual(query.query_dict, test_dict)

    def test_parse_element_strings(self):
        arg = "[VII][Fe,Ru,Os][I]"
        elements = parse_element_string(arg)
        self.assertEqual(elements, ["[VII]", "[Fe,Ru,Os]", "[I]"])

        arg = "[VII][Fe,Ru,Os][I][V][VIII][ASDASD]"
        elements = parse_element_string(arg)
        self.assertEqual(
            elements, ["[VII]", "[Fe,Ru,Os]", "[I]", "[V]", "[VIII]", "[ASDASD]"]
        )

        arg = "{VII}[Fe,Ru,Os][I]{V}[VIII][ASDASD]"
        elements = parse_element_string(arg)
        self.assertEqual(
            elements, ["{VII}", "[Fe,Ru,Os]", "[I]", "{V}", "[VIII]", "[ASDASD]"]
        )

        arg = "{VII}[Fe,Ru,Os][I][V][VIII][ASDASD]"
        elements = parse_element_string(arg)
        self.assertEqual(
            elements, ["{VII}", "[Fe,Ru,Os]", "[I]", "[V]", "[VIII]", "[ASDASD]"]
        )

        arg = "[VII]5[Fe,Ru,Os]2[I][V]6[VIII]2[ASDASD]"
        elements = parse_element_string(arg, stoich=True)
        self.assertEqual(
            elements,
            [
                "[VII]",
                "5",
                "[Fe,Ru,Os]",
                "2",
                "[I]",
                "[V]",
                "6",
                "[VIII]",
                "2",
                "[ASDASD]",
            ],
        )

        raised = False
        arg = "{VII}[Fe,Ru,Os}[I][V][VIII][ASDASD]"
        try:
            elements = parse_element_string(arg)
        except RuntimeError:
            raised = True
        self.assertTrue(raised, msg="Failed to raise error for unmatched brackets")

    def test_macro_composition(self):
        kwargs = {"composition": ["[I]FeBe"], "ignore_warnings": True, "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Na"]}},
                                {"elems": {"$in": ["K"]}},
                                {"elems": {"$in": ["Rb"]}},
                                {"elems": {"$in": ["Cs"]}},
                                {"elems": {"$in": ["Fr"]}},
                            ]
                        },
                        {"elems": {"$in": ["Fe"]}},
                        {"elems": {"$in": ["Be"]}},
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_intersection(self):
        kwargs = {
            "composition": ["LiFeBe"],
            "ignore_warnings": True,
            "intersection": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$or": [
                        {
                            "$and": [
                                {"stoichiometry": {"$size": 1}},
                                {"elems": {"$in": ["Li"]}},
                            ]
                        },
                        {
                            "$and": [
                                {"stoichiometry": {"$size": 1}},
                                {"elems": {"$in": ["Fe"]}},
                            ]
                        },
                        {
                            "$and": [
                                {"stoichiometry": {"$size": 1}},
                                {"elems": {"$in": ["Be"]}},
                            ]
                        },
                        {
                            "$and": [
                                {"stoichiometry": {"$size": 2}},
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Fe"]}},
                            ]
                        },
                        {
                            "$and": [
                                {"stoichiometry": {"$size": 2}},
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Be"]}},
                            ]
                        },
                        {
                            "$and": [
                                {"stoichiometry": {"$size": 2}},
                                {"elems": {"$in": ["Fe"]}},
                                {"elems": {"$in": ["Be"]}},
                            ]
                        },
                        {
                            "$and": [
                                {"stoichiometry": {"$size": 3}},
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Fe"]}},
                                {"elems": {"$in": ["Be"]}},
                            ]
                        },
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_middle_macro_composition(self):
        kwargs = {"composition": ["Fe[I]Be"], "ignore_warnings": True, "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["Fe"]}},
                        {
                            "$or": [
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Na"]}},
                                {"elems": {"$in": ["K"]}},
                                {"elems": {"$in": ["Rb"]}},
                                {"elems": {"$in": ["Cs"]}},
                                {"elems": {"$in": ["Fr"]}},
                            ]
                        },
                        {"elems": {"$in": ["Be"]}},
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_double_macro_composition(self):
        kwargs = {
            "composition": ["[Fe,Ru,Os][I]Be"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"elems": {"$in": ["Fe"]}},
                                {"elems": {"$in": ["Ru"]}},
                                {"elems": {"$in": ["Os"]}},
                            ]
                        },
                        {
                            "$or": [
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Na"]}},
                                {"elems": {"$in": ["K"]}},
                                {"elems": {"$in": ["Rb"]}},
                                {"elems": {"$in": ["Cs"]}},
                                {"elems": {"$in": ["Fr"]}},
                            ]
                        },
                        {"elems": {"$in": ["Be"]}},
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_double_end_macro_composition(self):
        kwargs = {
            "composition": ["Be[Fe,Ru,Os][I]"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["Be"]}},
                        {
                            "$or": [
                                {"elems": {"$in": ["Fe"]}},
                                {"elems": {"$in": ["Ru"]}},
                                {"elems": {"$in": ["Os"]}},
                            ]
                        },
                        {
                            "$or": [
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Na"]}},
                                {"elems": {"$in": ["K"]}},
                                {"elems": {"$in": ["Rb"]}},
                                {"elems": {"$in": ["Cs"]}},
                                {"elems": {"$in": ["Fr"]}},
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_triple_macro_composition(self):
        kwargs = {
            "composition": ["[VII][Fe,Ru,Os][I]"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"elems": {"$in": ["F"]}},
                                {"elems": {"$in": ["Cl"]}},
                                {"elems": {"$in": ["Br"]}},
                                {"elems": {"$in": ["I"]}},
                                {"elems": {"$in": ["At"]}},
                            ]
                        },
                        {
                            "$or": [
                                {"elems": {"$in": ["Fe"]}},
                                {"elems": {"$in": ["Ru"]}},
                                {"elems": {"$in": ["Os"]}},
                            ]
                        },
                        {
                            "$or": [
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Na"]}},
                                {"elems": {"$in": ["K"]}},
                                {"elems": {"$in": ["Rb"]}},
                                {"elems": {"$in": ["Cs"]}},
                                {"elems": {"$in": ["Fr"]}},
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        # print(json.dumps(test_dict, indent=2))
        # print(json.dumps(query.query_dict, indent=2))
        self.assertDictEqual(test_dict, query.query_dict)

    def test_double_list_composition(self):
        kwargs = {
            "composition": ["[Si,Ge,Sn][Fe,Ru,Os]"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"elems": {"$in": ["Si"]}},
                                {"elems": {"$in": ["Ge"]}},
                                {"elems": {"$in": ["Sn"]}},
                            ]
                        },
                        {
                            "$or": [
                                {"elems": {"$in": ["Fe"]}},
                                {"elems": {"$in": ["Ru"]}},
                                {"elems": {"$in": ["Os"]}},
                            ]
                        },
                        {"stoichiometry": {"$size": 2}},
                    ]
                }
            ]
        }

        self.assertDictEqual(test_dict, query.query_dict)

    def test_triple_list_composition(self):
        kwargs = {
            "composition": ["[Li,Na,K][Fe,Ru,Os][F,P,S]"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"elems": {"$in": ["Li"]}},
                                {"elems": {"$in": ["Na"]}},
                                {"elems": {"$in": ["K"]}},
                            ]
                        },
                        {
                            "$or": [
                                {"elems": {"$in": ["Fe"]}},
                                {"elems": {"$in": ["Ru"]}},
                                {"elems": {"$in": ["Os"]}},
                            ]
                        },
                        {
                            "$or": [
                                {"elems": {"$in": ["F"]}},
                                {"elems": {"$in": ["P"]}},
                                {"elems": {"$in": ["S"]}},
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        print(query.query_dict)
        self.assertDictEqual(test_dict, query.query_dict)

    def test_single_set_composition(self):
        kwargs = {"composition": ["{Li,Na}"], "ignore_warnings": True, "testing": True}
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"$and": [{"elems": {"$in": ["Li"]}}]},
                                {"$and": [{"elems": {"$in": ["Na"]}}]},
                            ]
                        },
                        {"stoichiometry": {"$size": 1}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_double_set_composition(self):
        kwargs = {
            "composition": ["{Li,Na}{Be,Fe}"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Li"]}},
                                        {"elems": {"$in": ["Be"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Li"]}},
                                        {"elems": {"$in": ["Fe"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Na"]}},
                                        {"elems": {"$in": ["Be"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Na"]}},
                                        {"elems": {"$in": ["Fe"]}},
                                    ]
                                },
                            ]
                        },
                        {"stoichiometry": {"$size": 2}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_ternary_set_composition(self):
        kwargs = {
            "composition": ["LiRu{Ru,S}"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Li"]}},
                                        {"elems": {"$in": ["Ru"]}},
                                        {"elems": {"$in": ["S"]}},
                                    ]
                                }
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_solo_set_composition(self):
        kwargs = {
            "composition": ["{Li}Ru{Ru,S}"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Li"]}},
                                        {"elems": {"$in": ["Ru"]}},
                                        {"elems": {"$in": ["S"]}},
                                    ]
                                }
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_triple_set_composition(self):
        kwargs = {
            "composition": ["{Li,Na}{Ru,Os}{Ru,S}"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Li"]}},
                                        {"elems": {"$in": ["Ru"]}},
                                        {"elems": {"$in": ["S"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Li"]}},
                                        {"elems": {"$in": ["Os"]}},
                                        {"elems": {"$in": ["Ru"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Li"]}},
                                        {"elems": {"$in": ["Os"]}},
                                        {"elems": {"$in": ["S"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Na"]}},
                                        {"elems": {"$in": ["Ru"]}},
                                        {"elems": {"$in": ["S"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Na"]}},
                                        {"elems": {"$in": ["Os"]}},
                                        {"elems": {"$in": ["Ru"]}},
                                    ]
                                },
                                {
                                    "$and": [
                                        {"elems": {"$in": ["Na"]}},
                                        {"elems": {"$in": ["Os"]}},
                                        {"elems": {"$in": ["S"]}},
                                    ]
                                },
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_illegal_character(self):
        kwargs = {
            "composition": ["{Li,Na,K}.{Fe,Ru,Os>}{F,P,S}"],
            "ignore_warnings": True,
            "testing": True,
        }
        with self.assertRaises(RuntimeError):
            DBQuery(**kwargs)

    def test_triple_macro_stoich(self):
        kwargs = {
            "formula": ["[VII]2[Fe,Ru,Os]3[I]"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"stoichiometry": {"$in": [["F", 2]]}},
                                {"stoichiometry": {"$in": [["Cl", 2]]}},
                                {"stoichiometry": {"$in": [["Br", 2]]}},
                                {"stoichiometry": {"$in": [["I", 2]]}},
                                {"stoichiometry": {"$in": [["At", 2]]}},
                            ]
                        },
                        {
                            "$or": [
                                {"stoichiometry": {"$in": [["Fe", 3]]}},
                                {"stoichiometry": {"$in": [["Ru", 3]]}},
                                {"stoichiometry": {"$in": [["Os", 3]]}},
                            ]
                        },
                        {
                            "$or": [
                                {"stoichiometry": {"$in": [["Li", 1]]}},
                                {"stoichiometry": {"$in": [["Na", 1]]}},
                                {"stoichiometry": {"$in": [["K", 1]]}},
                                {"stoichiometry": {"$in": [["Rb", 1]]}},
                                {"stoichiometry": {"$in": [["Cs", 1]]}},
                                {"stoichiometry": {"$in": [["Fr", 1]]}},
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        print(test_dict)
        print(query.query_dict)
        self.assertDictEqual(test_dict, query.query_dict)

        kwargs = {
            "formula": ["[Ag,Cd,In]2[Fe,Ru,Os]3[I]"],
            "ignore_warnings": True,
            "testing": True,
        }
        query = DBQuery(**kwargs)
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {
                            "$or": [
                                {"stoichiometry": {"$in": [["Ag", 2]]}},
                                {"stoichiometry": {"$in": [["Cd", 2]]}},
                                {"stoichiometry": {"$in": [["In", 2]]}},
                            ]
                        },
                        {
                            "$or": [
                                {"stoichiometry": {"$in": [["Fe", 3]]}},
                                {"stoichiometry": {"$in": [["Ru", 3]]}},
                                {"stoichiometry": {"$in": [["Os", 3]]}},
                            ]
                        },
                        {
                            "$or": [
                                {"stoichiometry": {"$in": [["Li", 1]]}},
                                {"stoichiometry": {"$in": [["Na", 1]]}},
                                {"stoichiometry": {"$in": [["K", 1]]}},
                                {"stoichiometry": {"$in": [["Rb", 1]]}},
                                {"stoichiometry": {"$in": [["Cs", 1]]}},
                                {"stoichiometry": {"$in": [["Fr", 1]]}},
                            ]
                        },
                        {"stoichiometry": {"$size": 3}},
                    ]
                }
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_query_time(self):
        from bson.objectid import ObjectId
        from datetime import datetime, timedelta
        from time import mktime

        num_days = 5
        kwargs = {"composition": "KP", "testing": True, "time": num_days}
        query = DBQuery(**kwargs)
        days_ago = (datetime.today() - timedelta(days=num_days)).timetuple()
        time_str = str(hex(int(mktime(days_ago))))[2:]
        time_str += (24 - len(time_str)) * "0"
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["K"]}},
                        {"elems": {"$in": ["P"]}},
                        {"stoichiometry": {"$size": 2}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
                {"_id": {"$lte": ObjectId(time_str)}},
            ]
        }
        print(test_dict["$and"][0])
        print(query.query_dict["$and"][0])
        self.assertDictEqual(test_dict, query.query_dict)

        num_days = 3001
        kwargs = {"composition": "KP", "testing": True, "time": num_days}
        query = DBQuery(**kwargs)
        days_ago = (datetime.today() - timedelta(days=num_days)).timetuple()
        time_str = str(hex(int(mktime(days_ago))))[2:]
        time_str += (24 - len(time_str)) * "0"
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["K"]}},
                        {"elems": {"$in": ["P"]}},
                        {"stoichiometry": {"$size": 2}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
                {"_id": {"$lte": ObjectId(time_str)}},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

        num_days = -1
        kwargs = {"composition": "KP", "testing": True, "time": num_days}
        query = DBQuery(**kwargs)
        days_ago = (datetime.today() - timedelta(days=num_days)).timetuple()
        time_str = str(hex(int(mktime(days_ago))))[2:]
        time_str += (24 - len(time_str)) * "0"
        test_dict = {
            "$and": [
                {
                    "$and": [
                        {"elems": {"$in": ["K"]}},
                        {"elems": {"$in": ["P"]}},
                        {"stoichiometry": {"$size": 2}},
                    ]
                },
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
                {"_id": {"$lte": ObjectId(time_str)}},
            ]
        }
        self.assertDictEqual(test_dict, query.query_dict)

    def test_id_query(self):
        kwargs = {"id": "testing testing", "ignore_warnings": False, "testing": True}
        query = DBQuery(**kwargs)

        test_dict = {
            "$and": [
                {"text_id": ["testing", "testing"]},
                {"$or": [{"quality": {"$gt": 0}}, {"quality": {"$exists": False}}]},
            ]
        }

        print(test_dict)
        print(query.query_dict)
        self.assertDictEqual(test_dict, query.query_dict)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import unittest


class ImportTest(unittest.TestCase):
    """ Test importing of all matador submodules. """

    def test_battery(self):
        from matador.battery import Electrode

        self.assertTrue(Electrode)

    def test_calculators(self):
        from matador.calculators import CastepCalculator

        self.assertTrue(CastepCalculator)

    def test_compute(self):
        from matador.compute import ComputeTask, BatchRun, reset_job_folder

        self.assertTrue(ComputeTask)
        self.assertTrue(BatchRun)
        self.assertTrue(reset_job_folder)

    def test_config(self):
        from matador.config import load_custom_settings

        self.assertTrue(load_custom_settings)

    def test_crystal(self):
        from matador.crystal import Crystal

        self.assertTrue(Crystal)

    def test_db(self):
        from matador.db import (
            make_connection_to_collection,
            Spatula,
            DatabaseChanges,
            Refiner,
        )

        self.assertTrue(make_connection_to_collection)
        self.assertTrue(Spatula)
        self.assertTrue(DatabaseChanges)
        self.assertTrue(Refiner)

    def test_export(self):
        from matador.export import (
            doc2param,
            doc2cell,
            doc2pdb,
            doc2pwscf,
            doc2res,
            doc2xsf,
            query2files,
            doc2arbitrary,
        )

        self.assertTrue(doc2param)
        self.assertTrue(doc2cell)
        self.assertTrue(doc2pdb)
        self.assertTrue(doc2pwscf)
        self.assertTrue(doc2res)
        self.assertTrue(doc2xsf)
        self.assertTrue(query2files)
        self.assertTrue(doc2arbitrary)

    def test_fingerprints(self):
        from matador.fingerprints import (
            PDF,
            PDFOverlap,
            CombinedProjectedPDF,
            PXRD,
            Fingerprint,
            FingerprintFactory,
            get_uniq_cursor,
        )

        self.assertTrue(
            all(
                [
                    PDF,
                    PDFOverlap,
                    CombinedProjectedPDF,
                    PXRD,
                    Fingerprint,
                    FingerprintFactory,
                    get_uniq_cursor,
                ]
            )
        )

    def test_hull(self):
        from matador.hull import (
            QueryConvexHull,
            EnsembleHull,
            PhaseDiagram,
            HullDiff,
            diff_hulls,
        )

        self.assertTrue(QueryConvexHull)
        self.assertTrue(EnsembleHull)
        self.assertTrue(PhaseDiagram)
        self.assertTrue(HullDiff)
        self.assertTrue(diff_hulls)

    def test_query(self):
        from matador.query import DBQuery

        self.assertTrue(DBQuery)

    def test_swaps(self):
        from matador.swaps import AtomicSwapper

        self.assertTrue(AtomicSwapper)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements an interface for querying the
__changelog_<collection> collections to allow for display and reversion
of particular database changes.

"""


from matador.db import make_connection_to_collection
from matador.utils.print_utils import print_warning, print_notify


class DatabaseChanges:
    """ Class to view and undo particular
    database changesets.

    """
    def __init__(self, collection_name: str, changeset_ind=0, action='view', override=False, mongo_settings=None):
        """ Parse arguments and run changes interface.

        Parameters:
            collection_name (str): the base collection name to act upon

        Keyword arguments:
            changset_ind (int): the number of the changset to act upon (1 is oldest)
            action (str): either 'view' or 'undo'
            override (bool): override all options to positive answers for testing
            mongo_settings (dict): dictionary of already-sources mongo settings

        """
        self.changelog_name = '__changelog_{}'.format(collection_name)
        _, _, self.collections = make_connection_to_collection(self.changelog_name,
                                                               allow_changelog=True,
                                                               override=override,
                                                               mongo_settings=mongo_settings)
        self.repo = [self.collections[key] for key in self.collections][0]
        curs = list(self.repo.find())

        if not curs:
            exit('No changesets found for {}'.format(collection_name))

        # if no changeset specified, print summary
        if changeset_ind == 0:
            self.print_change_summary(curs)

        elif changeset_ind > len(curs):
            exit('No changeset {} found for collection called "{}".'.format(changeset_ind, collection_name))

        # otherwise, try to act on particular changeset
        elif changeset_ind <= len(curs):
            self.change = curs[changeset_ind-1]
            self.view_changeset(self.change, changeset_ind-1)
            if action == 'undo':
                count = curs[changeset_ind-1]['count']
                print_warning('An attempt will now be made to remove {} structures from {}.'
                              .format(count, collection_name))
                print_notify('Are you sure you want to do that? (y/n)')
                if override:
                    response = 'y'
                else:
                    response = input()
                if response.lower() == 'y':
                    print_notify('You don\'t have any doubts at all? (y/n)')
                    if override:
                        next_response = 'n'
                    else:
                        next_response = input()
                    if next_response.lower() == 'n':
                        print('You\'re the boss, deleting structures now...')
                    else:
                        exit('As I thought...')
                else:
                    exit()

                # proceed with deletion
                _, _, collections = make_connection_to_collection(collection_name,
                                                                  allow_changelog=False,
                                                                  override=override)
                collection_to_delete_from = [collections[key] for key in collections][0]
                result = collection_to_delete_from.delete_many({'_id': {'$in': self.change['id_list']}})
                print('Deleted {}/{} successfully.'.format(result.deleted_count, self.change['count']))
                print('Tidying up changelog database...')
                self.repo.delete_one({'_id': self.change['_id']})
                if not self.repo.find_one():
                    print('No structures left remaining, deleting database...')
                    collection_to_delete_from.drop()
                    self.repo.drop()
                print('Success!')

    @staticmethod
    def view_changeset(changeset, index):
        """ Prints all details about a particular changeset.

        Parameters:
            changeset (dict): changeset stored in changelog database
            index (int): changeset index

        """
        print('Files added by changeset:')
        for src in changeset['src_list']:
            print('(+) {}'.format(src))
        print('({}) {} {:>7d} additions'.format(index+1, changeset['date'], changeset['count']))

    @staticmethod
    def print_change_summary(curs):
        """ Prints a summary of changes.

        Parameters:
            curs (list): cursor from changelog database

        """
        for index, change in enumerate(curs):
            print('({}) {} {:>7d} additions'.format(index+1, change['date'], change['count']))



#!/usr/bin/env python
import unittest
from matador.swaps import AtomicSwapper
from matador.utils.chem_utils import get_periodic_table


class SwapTest(unittest.TestCase):
    """ Test atomic swap functions. """

    def test_single_simple_swap(self):
        # spoof AtomicSwapper __init__
        swap_args = {"swap": ["AsP"], "debug": True}
        bare_swap = AtomicSwapper([])
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(bare_swap.swap_pairs, [[["As"], ["P"]]])
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["Li", "Li", "As", "As"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 1)
        self.assertEqual(swapped_docs[0]["atom_types"], ["Li", "Li", "P", "P"])

    def test_null_self_swap(self):
        # spoof AtomicSwapper __init__
        swap_args = {"swap": ["KK:PP"], "debug": True}
        bare_swap = AtomicSwapper([])
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(bare_swap.swap_pairs, [[["K"], ["K"]], [["P"], ["P"]]])
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["K", "K", "P", "P"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 0)

    def test_multiple_simple_swap(self):
        # spoof AtomicSwapper __init__
        swap_args = {"swap": ["AsP:LiNa"], "debug": True}
        bare_swap = AtomicSwapper([])
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(bare_swap.swap_pairs, [[["As"], ["P"]], [["Li"], ["Na"]]])
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["Li", "Li", "As", "As"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 1)
        self.assertEqual(swapped_docs[0]["atom_types"], ["Na", "Na", "P", "P"])

    def test_one_to_many_swap(self):
        # spoof AtomicSwapper __init__
        swap_args = {"swap": ["As[P,Sb,Zn,Cu]"], "debug": True}
        bare_swap = AtomicSwapper([])
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(bare_swap.swap_pairs, [[["As"], ["P", "Sb", "Zn", "Cu"]]])
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["Li", "Li", "As", "As"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 4)
        P_found = False
        Sb_found = False
        Zn_found = False
        Cu_found = False
        for new_doc in swapped_docs:
            self.assertTrue("As" not in new_doc["atom_types"])
            if "P" in new_doc["atom_types"]:
                self.assertTrue(
                    x not in new_doc["atom_types"] for x in ["Sb", "Zn", "Cu"]
                )
                self.assertEqual(new_doc["atom_types"], ["Li", "Li", "P", "P"])
                P_found = True
            if "Sb" in new_doc["atom_types"]:
                self.assertTrue(
                    x not in new_doc["atom_types"] for x in ["P", "Zn", "Cu"]
                )
                self.assertEqual(new_doc["atom_types"], ["Li", "Li", "Sb", "Sb"])
                Sb_found = True
            if "Zn" in new_doc["atom_types"]:
                self.assertTrue(
                    x not in new_doc["atom_types"] for x in ["P", "Sb", "Cu"]
                )
                self.assertEqual(new_doc["atom_types"], ["Li", "Li", "Zn", "Zn"])
                Zn_found = True
            if "Cu" in new_doc["atom_types"]:
                self.assertTrue(
                    x not in new_doc["atom_types"] for x in ["P", "Sb", "Zn"]
                )
                self.assertEqual(new_doc["atom_types"], ["Li", "Li", "Cu", "Cu"])
                Cu_found = True
        self.assertTrue(P_found)
        self.assertTrue(Sb_found)
        self.assertTrue(Zn_found)
        self.assertTrue(Cu_found)

    def test_mistaken_macro(self):
        swap_args = {"swap": ["VP"], "debug": True}
        bare_swap = AtomicSwapper([], maintain_num_species=False)
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(bare_swap.swap_pairs, [[["V"], ["P"]]])

    def test_many_to_one_macro(self):
        swap_args = {"swap": ["[V]P"], "debug": True}
        bare_swap = AtomicSwapper([], maintain_num_species=False)
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(bare_swap.swap_pairs, [[["N", "P", "As", "Sb", "Bi"], ["P"]]])
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["P", "Sb", "As", "As"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 1)
        self.assertEqual(swapped_docs[0]["atom_types"], ["P", "P", "P", "P"])

    def test_many_to_many_macro(self):
        swap_args = {"swap": ["[V][Tc,Mo]"], "debug": True}
        bare_swap = AtomicSwapper([], maintain_num_species=False)
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["P", "Sb", "As", "As", "Bi"]
        self.assertEqual(
            bare_swap.swap_pairs, [[["N", "P", "As", "Sb", "Bi"], ["Tc", "Mo"]]]
        )
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 2)
        self.assertEqual(swapped_docs[0]["atom_types"], ["Tc", "Tc", "Tc", "Tc", "Tc"])
        self.assertEqual(swapped_docs[1]["atom_types"], ["Mo", "Mo", "Mo", "Mo", "Mo"])

    def test_multiple_many_to_one_swap(self):
        # spoof AtomicSwapper __init__
        swap_args = {"swap": ["[Li,Na]K:[Ru, Rh]La"], "debug": True}
        bare_swap = AtomicSwapper([], maintain_num_species=False)
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(
            bare_swap.swap_pairs, [[["Li", "Na"], ["K"]], [["Ru", "Rh"], ["La"]]]
        )
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["Li", "Na", "Ru", "Rh"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 1)
        self.assertEqual(swapped_docs[0]["atom_types"], ["K", "K", "La", "La"])

    def test_many_to_many_swap_awkward(self):
        # spoof AtomicSwapper __init__
        swap_args = {"swap": ["[Li,Na]K:[V,I][I]"], "debug": True}
        bare_swap = AtomicSwapper([], maintain_num_species=False)
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        self.assertEqual(
            bare_swap.swap_pairs,
            [[["Li", "Na"], ["K"]], [["V", "I"], ["Li", "Na", "K", "Rb", "Cs", "Fr"]]],
        )
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["Li", "Na", "V", "I"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        self.assertEqual(num_swapped, 6)
        self.assertEqual(swapped_docs[0]["atom_types"], ["K", "K", "Li", "Li"])

    def test_maintain_num_species(self):
        # spoof AtomicSwapper __init__
        swap_args = {"swap": ["[Li,Na]K"], "debug": True}
        bare_swap = AtomicSwapper([])
        bare_swap.periodic_table = get_periodic_table()
        bare_swap.swap_args = swap_args["swap"]
        # try to parse swaps
        bare_swap.parse_swaps()
        # set up test data for real swap
        doc = dict()
        doc["atom_types"] = ["Li", "Na"]
        swapped_docs, num_swapped = bare_swap.atomic_swaps(doc)
        print(swapped_docs)
        self.assertEqual(num_swapped, 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

""" This file implements several tests that directly act on
a database instance running locally, more integration tests
than unittests per se.

"""


import unittest
import os
import sys
import glob

import mongomock

import matador.cli.cli
from matador.query import DBQuery
from matador.hull import QueryConvexHull
from matador.scrapers.castep_scrapers import cell2dict, res2dict
from matador.utils.chem_utils import get_root_source

REAL_PATH = "/".join(os.path.realpath(__file__).split("/")[:-1]) + "/"
CONFIG_FNAME = None
DB_NAME = "ci_test"
ROOT_DIR = os.getcwd()
DEBUG = False
MONGO_CLIENT = mongomock.MongoClient()
OUTPUT_DIR = ROOT_DIR + "/integration_test_results"


@mongomock.patch(servers=("mongo_test.com:27017",))
class IntegrationTest(unittest.TestCase):
    """ Test functionality acting on local database. """

    def tearDown(self):
        import shutil

        self.settings.reset()
        os.chdir(ROOT_DIR)
        shutil.rmtree(OUTPUT_DIR)

    def setUp(self):
        from matador.config import load_custom_settings, SETTINGS

        SETTINGS.reset()
        _ = load_custom_settings(config_fname=CONFIG_FNAME)
        SETTINGS["mongo"]["default_collection"] = DB_NAME
        SETTINGS["mongo"]["default_collection_file_path"] = "/data/"
        SETTINGS["mongo"]["host"] = "mongo_test.com"
        SETTINGS["mongo"]["port"] = 27017
        self.settings = SETTINGS

        os.makedirs(OUTPUT_DIR, exist_ok=False)
        os.chdir(OUTPUT_DIR)

    def test_integration(self):
        """ Test import and query. """
        print("IMPORT CASTEP 1")
        query = import_castep()
        self.assertEqual(
            len(query.cursor), 3, msg="Failed to import structures correctly"
        )

        # run again and hopefully nothing will change, i.e. no duplication
        print("IMPORT CASTEP 2")
        query = import_castep()
        self.assertEqual(
            len(query.cursor), 3, msg="Failed to import structures correctly"
        )

        print("IMPORT CASTEP 3")
        query = import_castep(extra_flags="--recent_only")
        self.assertEqual(
            len(query.cursor), 3, msg="Failed to import structures correctly"
        )

        print("IMPORT RES")
        query_1, query_2 = import_res()
        self.assertEqual(len(query_1.cursor), 7, msg="Failed to import res files")
        self.assertEqual(len(query_2.cursor), 4, msg="Failed to import res files")
        self.assertEqual(
            query_2.cursor[0]["species_pot"]["K"],
            "2|1.5|9|10|11|30U:40:31(qc=6)",
            msg="Failed to scrape OTF with weird name",
        )
        self.assertEqual(
            query_2.cursor[0]["species_pot"]["Sn"],
            "2|2|2|1.6|9.6|10.8|11.7|50U=-0.395U=+0.25:51U=-0.14U=+0.25",
            msg="Failed to scrape OTF with linebreak",
        )
        self.assertFalse(
            any(["Sb" in doc["species_pot"] for doc in query_2.cursor]),
            msg="pspots over-scraped!",
        )

        print("SWAP")
        output_folder_exists, successes, elem_successes = swaps()
        self.assertTrue(output_folder_exists, msg="No folder created")
        self.assertTrue(all(successes), msg="Failed to even read files")
        self.assertFalse(all(elem_successes), msg="Swaps had wrong elements")

        print("CHANGES")
        query_1, query_2, changes_count = changes()
        self.assertEqual(
            len(query_1.cursor), 3, msg="matador changes did not remove files"
        )
        self.assertEqual(
            len(query_2.cursor), 0, msg="matador changes did not remove files"
        )
        # unclear that mongomock can handle this; just check the queries instead
        # self.assertEqual(changes_count, 1, msg='matador changes did not changelog')

        expected_dir = "query-ci_test"
        if os.path.isdir(expected_dir):
            for _file in glob.glob(expected_dir + "/*"):
                os.remove(_file)
            os.removedirs(expected_dir)

        print("EXPORT")
        export()
        expected_files = [
            "query-ci_test/query-ci_test.md",
            "query-ci_test/query-ci_test.tex",
            "query-ci_test/Na3Zn4-swap-ReOs-OQMD_759599.pdb",
            "query-ci_test/Na3Zn4-swap-ReOs-OQMD_759599.xsf",
            "query-ci_test/Na3Zn4-swap-ReOs-OQMD_759599.json",
            "query-ci_test/Na-edgecase-CollCode10101.pdb",
            "query-ci_test/Na-edgecase-CollCode10101.xsf",
            "query-ci_test/Na-edgecase-CollCode10101.json",
            "query-ci_test/NaP_intermediates.pdb",
            "query-ci_test/NaP_intermediates.xsf",
            "query-ci_test/NaP_intermediates.json",
        ]
        dir_exists = os.path.isdir(expected_dir)
        files_exist = all([os.path.isfile(_file) for _file in expected_files])

        for _file in expected_files:
            if os.path.isfile(_file):
                os.remove(_file)

        if os.path.isdir(expected_dir):
            for _file in glob.glob(expected_dir + "/*"):
                os.remove(_file)
            os.removedirs(expected_dir)

        self.assertTrue(dir_exists, msg="Failed to create output directory")
        self.assertTrue(files_exist, msg="Some files missing from export")

        print("PSEUDOTERNARY HULL")
        query, hull = pseudoternary_hull()
        self.assertTrue(query.args.get("intersection"))
        self.assertTrue(query._non_elemental)
        self.assertTrue(query._create_hull)
        self.assertEqual(len(query.cursor), 7)
        self.assertEqual(len(hull.cursor), 7)
        self.assertEqual(len(hull.hull_cursor), 5)

        expected_dir = "query-LaLiZrO-ci_test"
        if os.path.isdir(expected_dir):
            for _file in glob.glob(expected_dir + "/*"):
                os.remove(_file)
            os.removedirs(expected_dir)

        print("PSEUDOTERNARY HULL 2")
        hull = pseudoternary_hull_no_query()
        self.assertTrue(hull._query.args.get("intersection"))
        self.assertTrue(hull._query._non_elemental)
        self.assertTrue(hull._query._create_hull)
        self.assertEqual(len(query.cursor), 7)
        self.assertEqual(len(hull.cursor), 7)
        self.assertEqual(len(hull.hull_cursor), 5)

        expected_dir = "query-LaLiZrO-ci_test"
        if os.path.isdir(expected_dir):
            for _file in glob.glob(expected_dir + "/*"):
                os.remove(_file)
            os.removedirs(expected_dir)

        print("UNIQ")
        uniq()
        expected_files = [expected_dir + "/cubic-LLZO-CollCode999999.res"]

        dir_exists = os.path.isdir(expected_dir)
        files_exist = all([os.path.isfile(_file) for _file in expected_files])
        correct_num = len(glob.glob(expected_dir + "/*.res"))

        if os.path.isdir(expected_dir):
            for _file in glob.glob(expected_dir + "/*"):
                os.remove(_file)
            os.removedirs(expected_dir)

        self.assertTrue(dir_exists, msg="Failed to create output directory, uniq")
        self.assertTrue(files_exist, msg="Some files missing from export, uniq")
        self.assertTrue(correct_num == len(expected_files), msg="Incorrect filter")

        print("STATS")
        try:
            stats()
        except TypeError:
            print("Unable to test stats module due to mongomock limitations.")
            pass

        print("ID QUERY")
        query = id_query()
        self.assertEqual(len(query.cursor), 0)

        print("REFINE")
        cursor = refine().cursor
        self.assertTrue(all([doc["doi"] == ["10/12345"] for doc in cursor]))
        self.assertTrue(all([doc["tags"] == ["integration_test"] for doc in cursor]))
        self.assertTrue(
            all([doc["root_source"] == get_root_source(doc) for doc in cursor])
        )
        self.assertTrue(all([isinstance(doc["_raw"], list) for doc in cursor]))


def import_castep(extra_flags=None):
    """ Import from castep files, returning data to be checked. """
    # import from CASTEP files only
    os.chdir(REAL_PATH + "/data/castep_files")
    sys.argv = ["matador", "import", "--force", "--db", DB_NAME]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    if DEBUG:
        sys.argv += ["--debug"]

    matador.cli.cli.main(no_quickstart=True)

    query = DBQuery(db=DB_NAME, config=CONFIG_FNAME, details=True, source=True)

    files_to_delete = glob.glob("*spatula*")
    for f in files_to_delete:
        os.remove(f)
    os.chdir(OUTPUT_DIR)

    return query


def import_res():
    """ Import from res files, returning data to be checked. """
    # import from combined res/cell/param files
    os.chdir(REAL_PATH + "/data/res_files")

    sys.argv = ["matador", "import", "--force", "--db", DB_NAME]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    if DEBUG:
        sys.argv += ["--debug"]

    matador.cli.cli.main(no_quickstart=True)

    query_1 = DBQuery(db=DB_NAME, config=CONFIG_FNAME, details=True, source=True)
    query_2 = DBQuery(
        db=DB_NAME, composition="KSnP", config=CONFIG_FNAME, details=True, source=True
    )

    files_to_delete = glob.glob("*spatula*")
    for f in files_to_delete:
        os.remove(f)

    os.chdir(OUTPUT_DIR)

    return query_1, query_2


def pseudoternary_hull():
    """ Import some other res files ready to make a hull. """
    os.chdir(REAL_PATH + "/data/hull-LLZO")
    sys.argv = ["matador", "import", "--force", "--db", DB_NAME]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    if DEBUG:
        sys.argv += ["--debug"]

    matador.cli.cli.main(no_quickstart=True)

    query = DBQuery(
        db=DB_NAME,
        subcmd="hull",
        composition="La2O3:Li2O:ZrO2",
        config=CONFIG_FNAME,
        details=True,
        source=True,
        no_plot=True,
    )
    hull = QueryConvexHull(query=query)

    os.chdir(OUTPUT_DIR)

    return query, hull


def pseudoternary_hull_no_query():
    """ Import some other res files ready to make a hull. """
    hull = QueryConvexHull(
        db=DB_NAME,
        composition="La2O3:Li2O:ZrO2",
        config=CONFIG_FNAME,
        details=True,
        source=True,
        no_plot=True,
    )

    os.chdir(OUTPUT_DIR)

    return hull


def stats():
    """ Run the stats command just to check it works at all. """
    sys.argv = ["matador", "stats", "--db", DB_NAME]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main()


def swaps():
    """ Run swaps, returning data to be checked. """
    sys.argv = [
        "matador",
        "swaps",
        "--db",
        DB_NAME,
        "--res",
        "--cell",
        "-c",
        "NaPZn",
        "-int",
        "-sw",
        "NaLi:PSi:ZnSn",
    ]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main()
    expected_dir = "swaps-NaPZn-ci_test-NaLi:PSi:ZnSn"
    output_folder_exists = os.path.isdir(expected_dir)
    successes = []
    elem_successes = []
    if output_folder_exists:
        os.chdir(expected_dir)
        expected_files = [
            "Li3Si2-swap-NaP_intermediates",
            "Li3Sn4-swap-Na3Zn4-ReOs-OQMD_759599",
            "Li-swap-Na-edgecase-CollCode10101",
        ]
        target_elems = ["Li", "Si", "Sn"]

        for files in expected_files:
            cell, s = cell2dict(files, db=False, positions=True)
            successes.append(s)
            if s:
                elems = set(cell["atom_types"])
                elem_successes.append(any([elem not in target_elems for elem in elems]))
            else:
                elem_successes.append(False)
            res, s = res2dict(files, db=False)
            successes.append(s)
            if s:
                elems = set(res["atom_types"])
                elem_successes.append(any([elem not in target_elems for elem in elems]))
            else:
                elem_successes.append(False)

    os.chdir(OUTPUT_DIR)

    return output_folder_exists, successes, elem_successes


def changes():
    """ Test matador changes functionality by undoing the second
    change (i.e. the res import).
    """
    sys.argv = ["matador", "changes", "--db", DB_NAME, "-c", "2", "--undo"]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    matador.cli.cli.main(no_quickstart=True)

    query_1 = DBQuery(db=DB_NAME, config=CONFIG_FNAME)
    query_2 = DBQuery(db=DB_NAME, composition="KSnP", config=CONFIG_FNAME)
    changes_count = MONGO_CLIENT.crystals["__changelog_" + DB_NAME].count_documents({})

    return query_1, query_2, changes_count


def refine():
    """ Test various matador refine tasks. """
    sys.argv = [
        "matador",
        "refine",
        "--db",
        DB_NAME,
        "--task",
        "sym",
        "--mode",
        "overwrite",
    ]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main(no_quickstart=True)

    sys.argv = [
        "matador",
        "refine",
        "--db",
        DB_NAME,
        "--task",
        "source",
        "--mode",
        "set",
    ]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main(no_quickstart=True)

    sys.argv = [
        "matador",
        "refine",
        "--db",
        DB_NAME,
        "--task",
        "doi",
        "--mode",
        "set",
        "--new_doi",
        "10/12345",
    ]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main(no_quickstart=True)

    sys.argv = [
        "matador",
        "refine",
        "--db",
        DB_NAME,
        "--task",
        "tag",
        "--mode",
        "overwrite",
        "--new_tag",
        "integration_test",
    ]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main(no_quickstart=True)

    sys.argv = [
        "matador",
        "refine",
        "--db",
        DB_NAME,
        "--task",
        "pspot",
        "--mode",
        "overwrite",
    ]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main(no_quickstart=True)

    sys.argv = [
        "matador",
        "refine",
        "--db",
        DB_NAME,
        "--task",
        "raw",
        "--mode",
        "overwrite",
    ]
    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]
    matador.cli.cli.main(no_quickstart=True)

    query = DBQuery(db=DB_NAME, config=CONFIG_FNAME)

    return query


def export():
    """ Test exporting to some random file types. Don't worry too much about
    the contents of the files yet, just that they exist with non-zero size.
    """
    sys.argv = [
        "matador",
        "query",
        "--db",
        DB_NAME,
        "--xsf",
        "--pdb",
        "--latex",
        "--json",
    ]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    matador.cli.cli.main(no_quickstart=True)


def uniq():
    """ Test filtering of structures by PDF. """
    sys.argv = ["matador", "query", "--db", DB_NAME, "-u 0.1", "--res", "-c", "LaLiZrO"]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    matador.cli.cli.main(no_quickstart=True)


def id_query():
    """ Test a simple ID query, and that nothing is found... """
    sys.argv = ["matador", "query", "--db", DB_NAME, "-i", "testing testing"]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    matador.cli.cli.main(no_quickstart=True)

    query = DBQuery(db=DB_NAME, config=CONFIG_FNAME, id="testing testing")
    return query


def drop(collname):
    """ Drop collection. """
    sys.argv = ["matador", "stats", "--delete", "--db", collname]

    if CONFIG_FNAME is not None:
        sys.argv += ["--config", CONFIG_FNAME]

    matador.cli.cli.main(no_quickstart=True)
    print("Dropped ci_test!")



#!/usr/bin/env python

""" These tests only check whether plots are created,
not that they look correct!

"""

import unittest
import os
import sys
from glob import glob
import numpy as np
import matador.cli.dispersion
from matador.scrapers import res2dict
from matador.hull import QueryConvexHull
from matador.plotting.battery_plotting import plot_voltage_curve
from matador.plotting.pdf_plotting import plot_pdf
from matador.plotting.pxrd_plotting import plot_pxrd
from .utils import MatadorUnitTest

REAL_PATH = "/".join(os.path.realpath(__file__).split("/")[:-1]) + "/"
ROOT_DIR = os.getcwd()

try:
    import matplotlib  # noqa
    matplotlib.use("Agg")

    MATPLOTLIB_PRESENT = True
except ImportError:
    MATPLOTLIB_PRESENT = False

try:
    import ternary # noqa

    TERNARY_PRESENT = True
except ImportError:
    TERNARY_PRESENT = False

try:
    import _tkinter  # noqa
except Exception:
    MATPLOTLIB_PRESENT = False


@unittest.skipIf(not MATPLOTLIB_PRESENT, "Skipping plotting tests.")
class SpectralPlotTests(unittest.TestCase):
    """ Test Dispersion script. """

    def test_pdis_plot(self):
        """ Test combined spectral plots. """
        os.chdir(REAL_PATH + "/data/dispersion")
        expected_file = "K3P-OQMD_4786-CollCode25550_spectral.png"
        if os.path.isfile(expected_file):
            os.remove(expected_file)
        sys.argv = [
            "dispersion",
            "K3P-OQMD_4786-CollCode25550",
            "--png",
            "-scale",
            "10",
            "-interp",
            "2",
            "-pw",
            "-5",
            "5",
            "--gap",
            "--preserve_kspace_distance",
            "--figsize",
            "10",
            "10",
        ]
        errored = False
        try:
            matador.cli.dispersion.main()
        except Exception as exc:
            errored = True
            error = exc
        file_exists = os.path.isfile(expected_file)
        if file_exists:
            os.remove(expected_file)
        os.chdir(ROOT_DIR)
        if errored:
            raise error
        self.assertTrue(file_exists)

    def test_dos_only(self):
        """ Test combined spectral plots. """
        os.chdir(REAL_PATH + "/data/dispersion")
        expected_file = "K3P-OQMD_4786-CollCode25550_spectral.png"
        if os.path.isfile(expected_file):
            os.remove(expected_file)
        sys.argv = [
            "dispersion",
            "K3P-OQMD_4786-CollCode25550",
            "--png",
            "--dos_only",
            "--figsize",
            "10",
            "10",
        ]
        errored = False
        try:
            matador.cli.dispersion.main()
        except Exception as exc:
            errored = True
            error = exc
        file_exists = os.path.isfile(expected_file)
        if file_exists:
            os.remove(expected_file)
        os.chdir(ROOT_DIR)
        if errored:
            raise error
        self.assertTrue(file_exists)

    def test_multiseed(self):
        """ Test plotting two seed bandstructures on top of each other. """
        os.chdir(REAL_PATH + "/data/bands_files")
        expected_file = "KPSn_spectral.png"
        sys.argv = [
            "dispersion",
            "KPSn",
            "KPSn_2.bands",
            "--dos_only",
            "--cmap",
            "viridis",
            "--png",
            "--band_reorder",
            "--labels",
            "PBE, LDA",
            "--figsize",
            "10",
            "10",
        ]
        errored = False
        try:
            matador.cli.dispersion.main()
        except Exception as exc:
            errored = True
            error = exc
        file_exists = os.path.isfile(expected_file)
        if file_exists:
            os.remove(expected_file)
        os.chdir(ROOT_DIR)
        if errored:
            raise error
        self.assertTrue(file_exists)

    def test_x11_no_fail(self):
        """ Test combined spectral plots. """
        os.chdir(REAL_PATH + "/data/dispersion")
        sys.argv = [
            "dispersion",
            "K3P-OQMD_4786-CollCode25550",
            "--dos_only",
            "--cmap",
            "viridis",
            "--figsize",
            "10",
            "10",
        ]
        errored = False
        try:
            matador.cli.dispersion.main()
        except Exception as exc:
            errored = True
            error = exc
        os.chdir(ROOT_DIR)
        if errored:
            raise error

    def test_phonon_dispersion(self):
        """ Test phonon dispersion plot. """
        os.chdir(REAL_PATH + "/data/phonon_dispersion")
        expected_file = "K3P_spectral.png"
        if os.path.isfile(expected_file):
            os.remove(expected_file)
        sys.argv = [
            "dispersion",
            "K3P",
            "--png",
            "-ph",
            "--band_colour",
            "k",
            "--figsize",
            "10",
            "10",
        ]
        errored = False
        try:
            matador.cli.dispersion.main()
        except Exception as exc:
            errored = True
            error = exc
        file_exists = os.path.isfile(expected_file)
        if file_exists:
            os.remove(expected_file)
        os.chdir(ROOT_DIR)
        if errored:
            raise error
        self.assertTrue(file_exists)

    def test_phonon_ir(self):
        """ Test phonon IR/Raman plot. """
        os.chdir(REAL_PATH + "/data/phonon_ir")
        expected_file = "h-BN_IRR_ir.svg"
        if os.path.isfile(expected_file):
            os.remove(expected_file)
        sys.argv = ["dispersion", "h-BN_IRR", "--svg", "-ir", "--figsize", "5", "5"]
        errored = False
        try:
            matador.cli.dispersion.main()
        except Exception as exc:
            errored = True
            error = exc
        file_exists = os.path.isfile(expected_file)
        if file_exists:
            os.remove(expected_file)
        os.chdir(ROOT_DIR)
        if errored:
            raise error
        self.assertTrue(file_exists)

    def test_projector_scraping(self):
        from matador.plotting.spectral_plotting import _parse_projectors_list
        self.assertEqual(
            _parse_projectors_list("K"),
            [("K", "s", None), ("K", "p", None), ("K", "d", None), ("K", "f", None), ("K", None, None)]
        )
        self.assertEqual(
            _parse_projectors_list("K,P"),
            [("K", "s", None), ("K", "p", None), ("K", "d", None), ("K", "f", None), ("K", None, None),
             ("P", "s", None), ("P", "p", None), ("P", "d", None), ("P", "f", None), ("P", None, None)]
        )
        self.assertEqual(
            _parse_projectors_list("K,P:s"),
            [("K", "s", None), ("K", "p", None), ("K", "d", None), ("K", "f", None), ("K", None, None),
             ("P", "s", None)]
        )
        self.assertEqual(
            _parse_projectors_list("123:x,P:s"),
            [("123", "x", None), ("P", "s", None)]
        )


@unittest.skipIf(not MATPLOTLIB_PRESENT, "Skipping plotting tests.")
class HullPlotTests(MatadorUnitTest):
    """ Tests for plotting phase diagrams. """

    def test_binary_hull_plot(self):
        """ Test plotting binary hull. """
        expected_files = ["KP_hull_simple.svg"]
        cursor = res2dict(REAL_PATH + "data/hull-KP-KSnP_pub/*.res")[0]
        QueryConvexHull(
            cursor=cursor,
            elements=["K", "P"],
            svg=True,
            hull_cutoff=0.0,
            plot_kwargs={"plot_fname": "KP_hull_simple", "svg": True},
        )
        for expected_file in expected_files:
            self.assertTrue(os.path.isfile(expected_file))

    def test_binary_battery_plots(self):
        """ Test plotting binary hull. """
        expected_files = ["KP_hull.png", "KP_voltage.png", "KP_volume.png"]
        cursor = res2dict(REAL_PATH + "data/hull-KP-KSnP_pub/*.res")[0]
        QueryConvexHull(
            cursor=cursor,
            elements=["K", "P"],
            no_plot=False,
            png=True,
            quiet=False,
            voltage=True,
            labels=True,
            label_cutoff=0.05,
            hull_cutoff=0.1,
            volume=True,
            plot_kwargs={"colour_by_source": True},
        )
        for expected_file in expected_files:
            self.assertTrue(os.path.isfile(expected_file))

    def test_voltage_labels(self):
        expected_files = ["KP_voltage.png"]
        cursor = res2dict(REAL_PATH + "data/hull-KP-KSnP_pub/*.res")[0]
        hull = QueryConvexHull(
            cursor=cursor, species="KP", no_plot=True, voltage=True, labels=True
        )

        plot_voltage_curve(hull.voltage_data, labels=True, savefig=expected_files[0])
        for expected_file in expected_files:
            self.assertTrue(os.path.isfile(expected_file))

    @unittest.skipIf(not TERNARY_PRESENT, "Skipping as python-ternary not found")
    def test_ternary_hull_plot(self):
        """ Test plotting ternary hull. """
        expected_files = ["KSnP_hull.png", "KSnP_voltage.png"]
        for expected_file in expected_files:
            if os.path.isfile(expected_file):
                os.remove(expected_file)
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        self.assertEqual(
            len(res_list),
            87,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        QueryConvexHull(
            cursor=cursor,
            elements=["K", "Sn", "P"],
            no_plot=False,
            png=True,
            quiet=False,
            voltage=True,
            labels=True,
            label_cutoff=0.05,
            hull_cutoff=0.1,
            capmap=True,
        )
        self.assertTrue(os.path.isfile(expected_file))
        for expected_file in expected_files:
            os.remove(expected_file)

    def test_beef_hull_plot(self):
        """ Test plotting BEEF hull. """
        from matador.hull import EnsembleHull
        from matador.scrapers import castep2dict

        expected_file = "KP_beef_hull.svg"
        if os.path.isfile(expected_file):
            os.remove(expected_file)

        cursor, s = castep2dict(REAL_PATH + "data/beef_files/*.castep", db=False)
        self.assertEqual(len(s), 0)

        beef_hull = EnsembleHull(
            cursor,
            "_beef",
            elements=["K", "P"],
            num_samples=10,
            energy_key="total_energy_per_atom",
            parameter_key="thetas",
        )

        beef_hull.plot_hull(svg=True)
        self.assertTrue(os.path.isfile(expected_file))
        os.remove(expected_file)


@unittest.skipIf(not MATPLOTLIB_PRESENT, "Skipping plotting tests.")
class FingerprintPlotTests(MatadorUnitTest):
    """ Test ability to plot PDF and PXRDs. """

    def test_pdf_plot(self):
        structure = res2dict(
            REAL_PATH + "data/res_files/KPSn-OQMD_123456.res", as_model=True
        )[0]
        plot_pdf(structure, png=True)
        self.assertTrue(os.path.isfile("K7PSn_pdf.png"))
        plot_pdf([structure, structure], filename="test_pdf", rmax=5, png=True)
        self.assertTrue(os.path.isfile("test_pdf.png"))

    def test_pxrd_plot(self):
        structure = res2dict(
            REAL_PATH + "data/res_files/KPSn-OQMD_123456.res", as_model=True
        )[0]
        plot_pxrd(structure, png=True)
        self.assertTrue(os.path.isfile("K7PSn_pxrd.png"))
        plot_pdf([structure, structure], filename="test_pxrd", png=True)
        self.assertTrue(os.path.isfile("test_pxrd.png"))


@unittest.skipIf(not MATPLOTLIB_PRESENT, "Skipping plotting tests.")
class ConvergencePlotTest(unittest.TestCase):
    """ Test the ability to read convergence data and make plots. """

    def setUp(self):
        os.chdir(REAL_PATH + "/data/convergence/")

    def tearDown(self):
        os.chdir(ROOT_DIR)

    def test_scraping_and_plotting(self):
        from matador.plotting.convergence_plotting import (
            get_convergence_data,
            get_convergence_files,
            combine_convergence_data,
            get_convergence_values,
        )
        from matador.plotting.convergence_plotting import plot_cutoff_kpt_grid

        kpt_files = get_convergence_files("completed_kpts")
        cutoff_files = get_convergence_files("completed_cutoff")
        kpt_data = get_convergence_data(
            kpt_files, conv_parameter="kpoints_mp_spacing", species=["Li"]
        )
        cutoff_data = get_convergence_data(
            cutoff_files, conv_parameter="cut_off_energy", species=["Li"]
        )
        data = combine_convergence_data(kpt_data, cutoff_data)
        self.assertEqual(
            data["Li-bcc"]["kpoints_mp_spacing"]["kpoints_mp_spacing"], [0.1, 0.07]
        )
        self.assertEqual(data["Li-bcc"]["cut_off_energy"]["cut_off_energy"], [300, 400])
        values, parameters = get_convergence_values(
            data["Li-bcc"], "cut_off_energy", "formation_energy_per_atom", log=True
        )
        self.assertEqual(parameters.tolist(), [300.0, 400.0])
        self.assertAlmostEqual(values.tolist()[0], 0.7291198427497395, places=6)
        self.assertEqual(values.tolist()[1], -np.inf)
        self.data = data

        expected_files = ["conv.svg"]
        for expected_file in expected_files:
            if os.path.isfile(expected_file):
                os.remove(expected_file)

        plot_cutoff_kpt_grid(self.data, svg=True)

        for file in expected_files:
            self.assertTrue(os.path.isfile(file))

        for expected_file in expected_files:
            os.remove(expected_file)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the ASECalculator class for use
with the compute module. It wraps simple calculators from
ASE for use with matador.

https://wiki.fysik.dtu.dk/ase/ase/calculators/calculators.html

"""

from matador.calculators.calculator import Calculator


class ASECalculator(Calculator):
    raise NotImplementedError



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The plotting module provides functions to make publication quality
plots for bandstructures, phase diagrams, densities of states, voltages
and volume curves.

Plot style settings can be configured by specifying a matplotlib style
(either built-in or a custom file). The default "matador" style can be
found in the matador.config submodule. User default styles can be set
with the plotting.default_style option in user's matadorrc. Minor tweaks
can be made by passing style dictionaries on a per plot level.

"""


__all__ = ['plot_spectral', 'plot_voltage_curve',
           'plot_volume_curve',
           'plot_2d_hull', 'plot_ternary_hull', 'plot_ir_spectrum',
           'get_linear_cmap', 'plot_free_energy', 'plot_cutoff_kpt_grid',
           'plot_ensemble_hull', 'plot_temperature_hull',
           'plot_pdf', 'plot_projected_pdf', 'plot_diff_overlap',
           'plot_projected_diff_overlap', 'plot_pxrd', 'set_style']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'

from matador.plotting.plotting import get_linear_cmap, set_style
from matador.plotting.spectral_plotting import plot_spectral
from matador.plotting.convergence_plotting import plot_cutoff_kpt_grid
from matador.plotting.battery_plotting import plot_voltage_curve, plot_volume_curve
from matador.plotting.hull_plotting import plot_2d_hull, plot_ternary_hull, plot_temperature_hull, plot_ensemble_hull
from matador.plotting.temperature_plotting import plot_free_energy
from matador.plotting.pdf_plotting import plot_pdf, plot_projected_pdf, plot_diff_overlap, plot_projected_diff_overlap
from matador.plotting.pxrd_plotting import plot_pxrd
from matador.plotting.ir_plotting import plot_ir_spectrum



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the CastepCalculator class for use
with the compute module.

"""

import os
from matador.calculators.calculator import Calculator
from matador.utils.errors import InputError, CalculationError

VALID_PSPOT_LIBS = ['C7', 'C8', 'C9', 'C17', 'C18', 'MS', 'HARD',
                    'QC5', 'NCP', 'NCP18', 'NCP17', 'NCP9']


class CastepCalculator(Calculator):
    """ The CASTEP calculator. """
    @staticmethod
    def verify_calculation_parameters(calculation_parameters, structure):
        errors = []
        if 'species_pot' not in calculation_parameters:
            calculation_parameters['species_pot'] = {'library': 'C18'}
        if 'library' not in calculation_parameters['species_pot']:
            for elem in structure['stoichiometry']:
                if elem[0] not in calculation_parameters['species_pot']:
                    msg = 'Unable to find pseudopotential specification for species {}'.format(elem[0])
                    errors.append(msg)
                elif ('|' not in calculation_parameters['species_pot'][elem[0]] and
                      not os.path.isfile(os.path.expanduser(calculation_parameters['species_pot'][elem[0]])) and
                      calculation_parameters['species_pot'][elem[0]] not in VALID_PSPOT_LIBS):
                    msg = ('Unable to find pseudopotential file/string/library: {}'
                           .format(calculation_parameters['species_pot'][elem[0]]))
                    errors.append(msg)

        if 'cut_off_energy' not in calculation_parameters and 'basis_precision' not in calculation_parameters:
            msg = 'Unable to find cut_off_energy field in param file'
            errors.append(msg)
        if 'xc_functional' not in calculation_parameters:
            msg = 'Unable to find xc_functional field in param file'
            errors.append(msg)
        if not any([string in calculation_parameters for string in ['kpoints_list', 'kpoints_mp_grid', 'kpoints_mp_spacing']]):
            msg = 'Unable to find kpoint specifications in cell'
            errors.append(msg)

        if errors:
            raise InputError('. '.join(errors))

    @staticmethod
    def verify_simulation_cell(structure):
        errors = []
        try:
            Calculator.verify_simulation_cell(structure)
        except CalculationError as exc:
            errors.append(str(exc))
        # any extra checks here

        if errors:
            raise CalculationError('. '.join(errors))

    def do_memcheck(self):
        return 0



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements all queries to the database, including parsing
user inputs, displaying results and calling other functionality.

"""


import sys
import random
from os import devnull
from itertools import combinations
from traceback import print_exc

import pymongo as pm
import numpy as np
from bson.json_util import dumps
from bson.objectid import ObjectId

from matador.utils.print_utils import print_warning, print_success, print_notify
from matador.utils.chem_utils import get_periodic_table
from matador.utils.chem_utils import parse_element_string, get_stoich_from_formula
from matador.utils.cursor_utils import display_results, filter_cursor_by_chempots
from matador.db import make_connection_to_collection
from matador.config import load_custom_settings


class DBQuery:
    """ Class that implements queries to MongoDB
    structure database.

    Attributes:
        cursor (list of dict or :obj:`pymongo.Cursor`): list or cursor of structures
            matching query.
        args (dict): contains all keyword arguments used to construct
            the query (see matador query --help) for list.
        query_dict (dict): dictionary passed to database for query
        calc_dict (dict): if performing a matching query (e.g. self.args['subcmd'] = 'hull'), this
            dictionary contains the parameters used to match to other structures
        repo (pymongo.collection.Collection): the pymongo collection that is being queried.
        top (int): number of structures to print/export set by self.args.get('top') (DEFAULT: 10).
        cursor_min_limit (int): if a query returns more structures than this, do not implicitly convert to a list.

    """

    # below this number of documents,
    # all queries will return a list rather than a pymongo Cursor
    cursor_min_limit = 1000

    def __init__(
        self,
        client=False,
        collections=False,
        subcmd='query',
        debug=False,
        quiet=False,
        mongo_settings=None,
        **kwargs
    ):
        """ Parse arguments from matador or API call before calling
        query.

        Keyword arguments:
            client (pm.MongoClient): the MongoClient to connect to.
            collections (dict of pm.collections.Collection): dictionary of pymongo Collections.
            subcmd (str): either 'query' or 'hull', 'voltage', 'hulldiff'.
                These will decide whether calcuation accuracies are matched
                in the final results.

        """
        # read args and set housekeeping
        self.args = kwargs
        self.debug = debug
        if self.args.get('subcmd') is None:
            self.args['subcmd'] = subcmd
        if self.args.get('testing') is None:
            self.args['testing'] = False
        if self.args.get('as_crystal') is None:
            self.args['as_crystal'] = False

        if subcmd in ['hull', 'hulldiff', 'voltage'] and self.args.get('composition') is None:
            raise RuntimeError('{} requires composition query'.format(subcmd))

        self._create_hull = (self.args.get('subcmd') in ['hull', 'hulldiff', 'voltage'] or
                             self.args.get('hull_cutoff') is not None)

        # public attributes
        self.cursor = EmptyCursor()
        self.query_dict = None
        self.calc_dict = None
        self.repo = None

        # private attributes to be set later
        self._empty_query = None
        self._gs_enthalpy = None
        self._non_elemental = None
        self._chempots = None
        self._num_to_display = None

        if debug:
            print(self.args)

        if quiet:
            f = open(devnull, 'w')
            sys.stdout = f

        # if testing keyword is used, all database operations are ignored
        if not self.args.get('testing'):

            # connect to db or use passed client
            if client:
                self._client = client
                self._db = client.crystals
            if collections is not False:
                _collections = collections

            if (not collections or not client):
                # use passed settings or load from config file
                if mongo_settings:
                    self.mongo_settings = mongo_settings
                else:
                    self.mongo_settings = load_custom_settings(
                        config_fname=self.args.get('config'), debug=self.args.get('debug')
                    )

                result = make_connection_to_collection(
                    self.args.get('db'), mongo_settings=self.mongo_settings
                )
                # ideally this would be rewritten to use a context manager to ensure
                # that connections are _always_ cleaned up
                self._client, self._db, _collections = result

            if len(_collections) > 1:
                raise NotImplementedError("Querying multiple collections is no longer supported.")
            else:
                for collection in _collections:
                    self._collection = _collections[collection]
                    break

        # define some periodic table macros
        self._periodic_table = get_periodic_table()

        # set default top value to 10
        if self.args.get('summary') or self.args.get('subcmd') in ['swaps', 'polish']:
            self.top = None
        else:
            self.top = self.args.get('top') if self.args.get('top') is not None else 10

        # create the dictionary to pass to MongoDB
        self._construct_query()

        if not self.args.get('testing'):

            if self.args.get('id') is not None and (self._create_hull or self.args.get('calc_match')):
                # if we've requested and ID and hull/calc_match, do the ID query
                self.perform_id_query()

            self.perform_query()

            if self._create_hull and self.args.get('id') is None:
                # if we're making a normal hull, find the sets of calculations to use
                self.perform_hull_query()

            if not self._create_hull:
                # only filter for uniqueness if not eventually making a hull
                if self.args.get('uniq'):
                    from matador.utils.cursor_utils import filter_unique_structures
                    print_notify('Filtering for unique structures...')

                    if isinstance(self.cursor, pm.cursor.Cursor):
                        raise RuntimeError("Unable to filter pymongo cursor for uniqueness directly.")

                    if self.args.get('top') is not None:
                        top = self.args['top']
                    else:
                        top = len(self.cursor)

                    self.cursor = filter_unique_structures(
                        self.cursor[:top],
                        debug=self.args.get('debug'),
                        sim_tol=self.args.get('uniq'),
                        energy_tol=1e20
                    )

            if self.args.get('available_values') is not None:
                print('Querying available values...')
                self._query_available_values(self.args.get('available_values'), self.cursor)

            # if no client was passed, then we need to close the one we made
            if not client and not self.args.get('testing'):
                self._client.close()

        if quiet:
            f.close()
            sys.stdout = sys.__stdout__

    def _construct_query(self):
        """ Set up query dict and perform query depending on
        command-line / API arguments. Sets self.query_dict.

        """
        self.cursor = EmptyCursor()
        # initalize query_dict to '$and' all queries
        self.query_dict = dict()
        self.query_dict['$and'] = []
        self._empty_query = True

        # benchmark enthalpy to display (set by calc_match)
        self._gs_enthalpy = 0.0

        # operate on one structure and related others
        if self.args.get('id') is not None:
            if not self._create_hull and not self.args.get('calc_match'):
                self.query_dict['$and'].append(self._query_id())
                self._empty_query = False

        # create alias for formula for backwards-compatibility
        self.args['stoichiometry'] = self.args.get('formula')
        if self.args.get('stoichiometry') is not None:
            self.query_dict['$and'].append(self._query_stoichiometry())
            self._empty_query = False

        if self.args.get('composition') is not None:
            self.query_dict['$and'].append(self._query_composition())
            self._empty_query = False

        if self.args.get('num_species') is not None:
            self.query_dict['$and'].append(self._query_num_species())
            self._empty_query = False

        if self.args.get('space_group') is not None:
            self.query_dict['$and'].append(self._query_space_group())
            self._empty_query = False

        if self.args.get('num_fu') is not None:
            self.query_dict['$and'].append(self._query_num_fu())
            self._empty_query = False

        if self.args.get('tags') is not None:
            self.query_dict['$and'].append(self._query_tags())
            self._empty_query = False

        if self.args.get('doi') is not None:
            self.query_dict['$and'].append(self._query_doi())
            self._empty_query = False

        if self.args.get('icsd') is not None:
            self.query_dict['$and'].append(self._query_icsd())
            self._empty_query = False

        if self.args.get('field') is not None:
            try:
                for ind, field in enumerate(self.args.get('field')):
                    _filter = self.args.get('filter')[ind]
                    try:
                        for i, value in enumerate(_filter):
                            _filter[i] = float(value)
                        filter_type = 'float'
                    except ValueError:
                        filter_type = 'string'

                    if filter_type == 'float':
                        self.query_dict['$and'].append(self._query_float_range(
                            field, _filter))
                    else:
                        self.query_dict['$and'].append(self._query_string(
                            field, _filter))
            except Exception:
                raise RuntimeError(
                    "Unexpected field/filter format. Both must be "
                    "provided as lists, even if only one field is being filtered."
                )
            self._empty_query = False

        if self.args.get('cutoff') is not None:
            self.query_dict['$and'].append(self._query_float_range(
                'cut_off_energy', self.args.get('cutoff')))
            self._empty_query = False

        if self.args.get('geom_force_tol') is not None:
            self.query_dict['$and'].append(self._query_float_range(
                'geom_force_tol', self.args.get('geom_force_tol')))
            self._empty_query = False

        if self.args.get('grid_scale') is not None:
            self.query_dict['$and'].append(self._query_float_range(
                'grid_scale', self.args.get('grid_scale')))
            self._empty_query = False

        if self.args.get('fine_grid_scale') is not None:
            self.query_dict['$and'].append(self._query_float_range(
                'fine_grid_scale', self.args.get('fine_grid_scale')))
            self._empty_query = False

        if self.args.get('src_str') is not None:
            self.query_dict['$and'].append(self._query_source())
            self._empty_query = False

        if self.args.get('root_src') is not None:
            self.query_dict['$and'].append(self._query_root_source())
            self._empty_query = False

        if self.args.get('pressure') is not None:
            self.query_dict['$and'].append(self._query_float_range(
                'pressure', self.args.get('pressure') or 0.0, tolerance=self.args.get('pressure_tolerance') or 0.5))
            self._empty_query = False

        elif self.args['subcmd'] in ['hull', 'hulldiff', 'voltage']:
            self.query_dict['$and'].append(self._query_float_range(
                'pressure', 0.0, tolerance=self.args.get('pressure_tolerance') or 0.5))

        if self.args.get('encapsulated') is True:
            self.query_dict['$and'].append(self._query_encap())
            self._empty_query = False

        if self.args.get('cnt_radius') is not None:
            self.query_dict['$and'].append(self._query_float_range(
                'cnt_radius', self.args.get('cnt_radius'), tolerance=0.01))
            self._empty_query = False

        if self.args.get('cnt_vector') is not None:
            self.query_dict['$and'].append(self._query_cnt_vector())
            self._empty_query = False

        if self.args.get('sedc') is not None:
            self.query_dict['$and'].append(self._query_sedc())
            self._empty_query = False

        if self.args.get('xc_functional') is not None:
            self.query_dict['$and'].append(self._query_xc_functional())
            self._empty_query = False

        if self.args.get('mp_spacing') is not None:
            self.query_dict['$and'].append(self._query_float_range(
                'kpoints_mp_spacing', self.args.get('mp_spacing'), tolerance=self.args.get('kpoint_tolerance') or 0.01))
            self._empty_query = False

        if self.args.get('spin') is not None:
            tmp_dict = self._query_spin()
            if tmp_dict:
                self.query_dict['$and'].append(tmp_dict)
            self._empty_query = False

        if not self.args.get('ignore_warnings'):
            self.query_dict['$and'].append(self._query_quality())

        if self.args.get('time') is not None:
            self.query_dict['$and'].append(self._query_time(self.args.get('since') or False))
            self._empty_query = False

    def _perform_empty_query(self, as_list=False):
        """ No parameters were asked for, so just return a cursor
        that contains the entire collection currently stored as
        :attr:`repo`.

        Returns:
            list or :obj:`pymongo.cursor.Cursor`: the results of the query.


        """
        num_documents = self.repo.count_documents({})
        cursor = self.repo.find().sort('enthalpy_per_atom', pm.ASCENDING)
        if self.debug:
            print('Empty query, showing all...')

        if num_documents < self.cursor_min_limit or as_list:
            return list(cursor), num_documents

        return cursor, num_documents

    def perform_query(self):
        """ Find results that match the query_dict
        inside the MongoDB database.
        """
        # if no query submitted, find all
        if self._empty_query and self.args.get('id') is None:
            self.repo = self._collection
            self.cursor, cursor_count = self._perform_empty_query()

        # if no special query has been made already, begin executing the query
        if not self._empty_query:
            self.repo = self._collection
            if self.debug:
                print('Query dict:')
                print(dumps(self.query_dict, indent=1))

            # execute query
            self.cursor, cursor_count = self._find_and_sort(self.query_dict)
            if self._non_elemental:
                self.cursor = filter_cursor_by_chempots(self._chempots, self.cursor)

        print('{} results found for query in {}.'.format(cursor_count, self.repo.name))
        self._num_to_display = cursor_count
        if self.args.get('subcmd') != 'swaps' and not self._create_hull:
            self._set_filter_display_results(cursor_count)

        # if a summary has been requested, cursor must be converted to list
        if self.args.get('summary'):
            self.cursor = list(self.cursor)

        if self.args.get('subcmd') != 'swaps' and not self._create_hull:
            if self._num_to_display >= 1 or self._num_to_display is None:
                if self._num_to_display == cursor_count:
                    display_results(self.cursor, **self.args)
                else:
                    display_results(self.cursor[:self._num_to_display], **self.args)

        if isinstance(self.cursor, pm.cursor.Cursor):
            self.cursor.rewind()

    def _set_filter_display_results(self, cursor_count):
        """ Filter and display the results based on the command line parameters. """
        # by default, show the top structures only
        # if delta_E requested, count how many exist below that energy
        if self.args.get('delta_E') is not None:
            if isinstance(self.cursor, pm.cursor.Cursor) and len(self.cursor.distinct('stoichiometry')) > 1:
                print('Multiple stoichiometries in cursor, unable to filter by energy with --delta_E.')
            else:
                self.cursor = list(self.cursor)
                gs_enthalpy = self.cursor[0]['enthalpy_per_atom']
                for ind, doc in enumerate(self.cursor[1:]):
                    if abs(doc['enthalpy_per_atom'] - gs_enthalpy) > self.args.get('delta_E'):
                        self._num_to_display = ind + 1
                        break

        elif self.top == -1 or self.top is None or cursor_count <= self.top:
            self._num_to_display = cursor_count
            self.top = cursor_count
        elif cursor_count > self.top:
            self._num_to_display = self.top

    def _find_and_sort(self, query_filter=None, as_list=False, **kwargs):
        """ Query `self.repo` using Pymongo arguments/kwargs. Sorts based
        on enthalpy_per_atom and optionally returns list of Crystals.

        Keyword arguments:
            query_filter (dict): the query to use. If None, perform a blank query.
            as_list (bool): whether to return a list of a pm.cursor.Cursor object.

        Returns:
            list/pm.cursor.Cursor: the results of the query.
            int: the number of results in the query.

        """
        from matador.crystal import Crystal
        if query_filter is None:
            query_filter = {}
        count = self.repo.count_documents(query_filter, **kwargs)
        cursor = self.repo.find(query_filter, **kwargs).sort('enthalpy_per_atom', pm.ASCENDING)

        if self.args.get('as_crystal'):
            return [Crystal(doc) for doc in cursor], count
        if count < self.cursor_min_limit or as_list:
            return list(cursor), count

        return cursor, count

    def perform_hull_query(self):
        """ Perform the multiple queries necessary to find possible
        calculation sets to create a convex hull from.

        Raises:
            SystemExit: if no structures are found for hull.

        """
        if self._collection is not None:
            self.repo = self._collection
            print('Creating hull from structures in query results.')
            if self.args.get('biggest'):
                print('\nFinding biggest calculation set for hull...\n')
            else:
                print('\nFinding the best calculation set for hull...')

            test_cursors = []
            test_cursor_count = []
            text_ids = []
            calc_dicts = []
            cutoff = []

            num_sample = 2
            num_rand_sample = 5 if self.args.get('biggest') else 3

            if isinstance(self.cursor, pm.cursor.Cursor):
                count = self.cursor.count()
            else:
                count = len(self.cursor)

            if count <= 0:
                raise SystemExit('No structures found for hull.')

            # generate some random indices to match to, make sure they are in order
            # so can be accessed without cursor rewinds
            sampling_indices = list(range(num_sample)) + sorted(random.sample(range(2, count), num_rand_sample))

            for ind in sampling_indices:
                doc = self.cursor[ind]
                text_ids.append(doc['text_id'])
                try:
                    self.query_dict = self._query_calc(doc)
                    cutoff.append(doc['cut_off_energy'])
                    calc_dicts.append(dict())
                    calc_dicts[-1]['$and'] = list(self.query_dict['$and'])
                    self.query_dict['$and'].append(self._query_composition())
                    if not self.args.get('ignore_warnings'):
                        self.query_dict['$and'].append(self._query_quality())

                    probe_cursor, probe_count = self._find_and_sort(self.query_dict)

                    if self._non_elemental:
                        probe_cursor = filter_cursor_by_chempots(self._chempots, probe_cursor)
                        probe_count = len(probe_cursor)

                    test_cursors.append(probe_cursor)
                    test_cursor_count.append(probe_count)

                    print("{:^24}: matched {} structures."
                          .format(' '.join(doc['text_id']), probe_count),
                          end='\t-> ')
                    print('{spin}{sedc}{functional} {cutoff} eV, {geom_force_tol} eV/A, {kpoints} 1/A.'
                          .format(spin="S-" if doc.get('spin_polarized') else '',
                                  sedc="+" + doc.get('sedc') + "+" if doc.get('sedc') else "",
                                  functional=doc["xc_functional"],
                                  cutoff=doc["cut_off_energy"],
                                  geom_force_tol=doc.get('geom_force_tol', 'xxx'),
                                  kpoints=doc.get('kpoints_mp_spacing', 'xxx')))

                    if test_cursor_count[-1] == count:
                        print('Matched all structures...')
                        break
                    if test_cursor_count[-1] > 2 * int(count / 3):
                        print('Matched at least 2/3 of total number, composing hull...')
                        break

                except Exception:
                    print_exc()
                    print_warning('Error with {}'.format(' '.join(doc['text_id'])))

            if self.args.get('biggest'):
                choice = np.argmax(np.asarray(test_cursor_count))
            else:
                # by default, find highest cutoff hull as first proxy for quality
                choice = np.argmax(np.asarray(cutoff))

            text_id = text_ids[choice]
            self.cursor = test_cursors[choice]
            self.calc_dict = calc_dicts[choice]
            if not test_cursor_count[choice]:
                raise RuntimeError('No structures found that match chemical potentials.')

            print_success('Composing hull from set containing {}'.format(' '.join(text_id)))

    def perform_id_query(self):
        """ Query the `text_id` field for the ID provided in the args for a calc_match
        or hull/voltage query. Use the results of the text_id query to match to other
        entries that have the same calculation parameters. Sets self.query_dict and
        self.calc_dict.

        Raises:
            RuntimeError: if no structures are found.

        """

        self.cursor = []
        query_dict = dict()
        query_dict['$and'] = []
        query_dict['$and'].append(self._query_id())
        if not self.args.get('ignore_warnings'):
            query_dict['$and'].append(self._query_quality())
        self.repo = self._collection
        self.cursor = list(self._find_and_sort(query_dict))

        if not self.cursor:
            raise RuntimeError('Could not find a match with {} try widening your search.'.format(self.args.get('id')))

        if len(self.cursor) >= 1:
            display_results(list(self.cursor)[:self.top], **self.args)

            if len(self.cursor) > 1:
                print_warning('Matched multiple structures with same text_id. The first one will be used.')

            # save special copy of calc_dict for hulls
            self.calc_dict = dict()
            self.calc_dict['$and'] = []
            # to avoid deep recursion, and since this is always called first
            # don't append, just set
            self.query_dict = self._query_calc(self.cursor[0])
            if self.args.get('composition'):
                self.args['intersection'] = True
                self.query_dict['$and'].append(self._query_composition())
            self.calc_dict['$and'] = list(self.query_dict['$and'])

    def query_stoichiometry(self, **kwargs):
        """ Alias for private function of the same name. """
        return self._query_stoichiometry(**kwargs)

    def query_composition(self, **kwargs):
        """ Alias for private function of the same name. """
        return self._query_composition(**kwargs)

    def query_tags(self, **kwargs):
        """ Alias for private function of the same name. """
        return self._query_tags(**kwargs)

    def query_quality(self, **kwargs):
        """ Alias for private function of the same name. """
        return self._query_quality(**kwargs)

    @staticmethod
    def _query_float_range(field, values, tolerance=None):
        """ Query all entries with field between float value range,
        or with float value.

        Parameters:
            field (str): the field to query.
            values (float/list of float): either single value, or list
                of 2 floats.

        Keyword arguments:
            tolerance (float): tolerance to add and subtract if single value is provided.

        Returns:
            dict: the constructed query.

        """
        query_dict = dict()
        query_dict[field] = dict()
        if not isinstance(values, list):
            values = [values]
        if len(values) == 2:
            if values[0] > values[1]:
                tmp = values[0]
                values[0] = values[1]
                values[1] = tmp

            query_dict[field]['$gte'] = values[0]
            query_dict[field]['$lte'] = values[1]
        else:
            if tolerance is None:
                query_dict[field]['$eq'] = values[0]
            else:
                query_dict[field]['$gte'] = round(values[0] - tolerance, 8)
                query_dict[field]['$lte'] = round(values[0] + tolerance, 8)
        return query_dict

    @staticmethod
    def _query_string(field, values):
        """ Query all entries for an exact string match on field.

        Parameters:
            field (str): the field to query.
            values (list or str): strings to query with $or joins.

        Returns:
            dict: the constructed query.

        """
        query_dict = dict()
        if not isinstance(values, list):
            values = [values]
        if len(values) > 1:
            query_dict['$or'] = []
            for value in values:
                query_dict['$or'].append({field: value})
        else:
            query_dict[field] = values[0]

        return query_dict

    def _query_stoichiometry(self, custom_stoich=None, partial_formula=None):
        """ Query DB for particular stoichiometry. """
        # alias stoichiometry
        if custom_stoich is None:
            stoich = self.args.get('stoichiometry')
            if isinstance(stoich, str):
                stoich = [stoich]
        else:
            stoich = custom_stoich
        if partial_formula is None:
            partial_formula = self.args.get('partial_formula')
        if ':' in stoich[0]:
            raise RuntimeError('Formula cannot contain ":", you probably meant to query composition.')

        stoich = get_stoich_from_formula(stoich[0], sort=False)

        query_dict = dict()
        query_dict['$and'] = []

        for ind, _ in enumerate(stoich):
            elem = stoich[ind][0]
            fraction = int(stoich[ind][1])

            if '[' in elem or ']' in elem:
                types_dict = dict()
                types_dict['$or'] = list()
                elem = elem.strip('[').strip(']')
                if elem in self._periodic_table:
                    for group_elem in self._periodic_table[elem]:
                        types_dict['$or'].append(dict())
                        types_dict['$or'][-1]['stoichiometry'] = dict()
                        types_dict['$or'][-1]['stoichiometry']['$in'] = [[group_elem, fraction]]
                    query_dict['$and'].append(types_dict)
                elif ',' in elem:
                    for group_elem in elem.split(','):
                        types_dict['$or'].append(dict())
                        types_dict['$or'][-1]['stoichiometry'] = dict()
                        types_dict['$or'][-1]['stoichiometry']['$in'] = [[group_elem, fraction]]
                    query_dict['$and'].append(types_dict)
            else:
                stoich_dict = dict()
                stoich_dict['stoichiometry'] = dict()
                stoich_dict['stoichiometry']['$in'] = [[elem, fraction]]
                query_dict['$and'].append(stoich_dict)
        if not partial_formula:
            size_dict = dict()
            size_dict['stoichiometry'] = dict()
            size_dict['stoichiometry']['$size'] = len(stoich)
            query_dict['$and'].append(size_dict)

        return query_dict

    @staticmethod
    def _query_ratio(ratios):
        """ Query DB for ratio of two elements. Ratios must be integers.

        Parameters:
            ratios (list): e.g.  ratios = [['MoS', 2], ['LiS', 1]]

        """
        query_dict = dict()
        for pair in ratios:
            query_dict['ratios.' + pair[0]] = int(pair[1])
        return query_dict

    def _query_composition(self, custom_elem=None, partial_formula=None, elem_field='elems'):
        """ Query DB for all structures containing all the elements
        taken as input. Passing this function a number is a deprecated
        feature, replaced by query_num_species.

        Keyword arguments:
            custom_elem (str): use to query custom string, rather than CLI args
            partial_formula (bool): remove stoich size from query if True
            elem_field (str): which field to query for elems, either `atom_types` or `elems`

        Returns:
            dict: dictionary containing database query.

        """
        if custom_elem is None:
            if isinstance(self.args.get('composition'), str):
                elements = [self.args.get('composition')]
            else:
                elements = self.args.get('composition')
        else:
            elements = custom_elem
        if partial_formula is None:
            partial_formula = self.args.get('partial_formula')

        self._non_elemental = False
        if ':' in elements[0]:
            self._non_elemental = True
            self.args['intersection'] = True
            self._chempots = elements[0].split(':')
            elements = [parse_element_string(elem) for elem in self._chempots]
            elements = list(dict.fromkeys([char for elem in elements for char in elem if char.isalpha()]))
        # if there's only one string, try split it by caps
        if not self._non_elemental:
            for char in elements[0]:
                if char.isdigit():
                    raise SystemExit('Composition cannot contain a number.')
            elements = parse_element_string(elements[0])

        or_preference = False
        for _, elem in enumerate(elements):
            if '{' in elem or '}' in elem:
                or_preference = True

        elements_tmp = [element for ind, element in enumerate(elements)
                        if element not in elements[:ind]]
        if len(elements_tmp) < len(elements):
            print('Ignoring duplicate element...')
        elements = elements_tmp

        if self.args.get('intersection'):
            if or_preference:
                raise RuntimeError('Intersection not implemented for overlapping sets, e.g. {}')

            query_dict = dict()
            query_dict['$or'] = []
            size = len(elements)
            # iterate over all combinations, limited by num species
            if self.args.get('num_species'):
                max_num = self.args.get('num_species')
                min_num = max_num
            else:
                max_num = 8
                min_num = 1
                if len(elements) > max_num:
                    print('Limiting query to up to {} elements per structure...'.format(max_num))
            for rlen in range(min_num, max_num+1):
                for combi in combinations(elements, r=rlen):
                    list_combi = list(combi)
                    types_dict = dict()
                    types_dict['$and'] = list()
                    types_dict['$and'].append(dict())
                    types_dict['$and'][-1]['stoichiometry'] = dict()
                    types_dict['$and'][-1]['stoichiometry']['$size'] = len(list_combi)
                    for elem in list_combi:
                        types_dict['$and'].append(dict())
                        types_dict['$and'][-1][elem_field] = dict()
                        types_dict['$and'][-1][elem_field]['$in'] = [elem]
                    query_dict['$or'].append(types_dict)
        else:
            # expand group macros
            query_dict = dict()
            query_dict['$and'] = []
            size = len(elements)

            if or_preference:
                element_slots = []
                for elem in elements:
                    if '[' in elem or '{' in elem:
                        elem = elem.strip('{').strip('}').strip('[').strip(']')
                        if elem in self._periodic_table:
                            element_slots.append(self._periodic_table[elem])
                        elif ',' in elem:
                            element_slots.append(elem.split(','))
                        else:
                            element_slots.append([elem])
                    else:
                        element_slots.append([elem])

                from itertools import product
                slots = [list(config) for config in product(*element_slots)]
                types_dict = dict()
                types_dict['$or'] = list()
                for slot in slots:
                    if len({elem for elem in slot}) == len(slot):
                        types_dict['$or'].append(dict())
                        types_dict['$or'][-1]['$and'] = []
                        for elem in slot:
                            types_dict['$or'][-1]['$and'].append(dict())
                            types_dict['$or'][-1]['$and'][-1][elem_field] = dict()
                            types_dict['$or'][-1]['$and'][-1][elem_field]['$in'] = [elem]
                query_dict['$and'].append(types_dict)

            else:
                for elem in elements:
                    if '[' in elem or ']' in elem:
                        types_dict = dict()
                        types_dict['$or'] = list()
                        elem = elem.strip('[').strip(']')
                        if elem in self._periodic_table:
                            for group_elem in self._periodic_table[elem]:
                                types_dict['$or'].append(dict())
                                types_dict['$or'][-1][elem_field] = dict()
                                types_dict['$or'][-1][elem_field]['$in'] = [group_elem]
                        elif ',' in elem:
                            for group_elem in elem.split(','):
                                types_dict['$or'].append(dict())
                                types_dict['$or'][-1][elem_field] = dict()
                                types_dict['$or'][-1][elem_field]['$in'] = [group_elem]
                    else:
                        types_dict = dict()
                        types_dict[elem_field] = dict()
                        types_dict[elem_field]['$in'] = [elem]

                    query_dict['$and'].append(types_dict)

        if not partial_formula and not self.args.get('intersection'):
            size_dict = dict()
            size_dict['stoichiometry'] = dict()
            size_dict['stoichiometry']['$size'] = size
            query_dict['$and'].append(size_dict)

        return query_dict

    def _query_num_species(self):
        """ Query database for all structures with a
        given number of elements, e.g. binaries, ternaries etc.
        """
        num = self.args.get('num_species')
        if not isinstance(num, list):
            num = num
        elif isinstance(num, list):
            num = num[0]
        else:
            sys.exit('--num_species takes a single integer or list containing a single integer')
        query_dict = dict()
        query_dict['stoichiometry'] = dict()
        query_dict['stoichiometry']['$size'] = num

        return query_dict

    def _query_space_group(self):
        """ Query DB for all structures with given
        space group.
        """
        query_dict = dict()
        if not isinstance(self.args.get('space_group'), list):
            spg = [self.args.get('space_group')]
        else:
            spg = self.args.get('space_group')
        query_dict['space_group'] = str(spg[0])

        return query_dict

    def _query_num_fu(self):
        """ Query DB for all structures with more than a
        given number of formula units in the simulation.
        """
        query_dict = dict()
        num = self.args.get('num_fu')
        if isinstance(num, list):
            num = num[0]
        query_dict['num_fu'] = dict()
        query_dict['num_fu']['$gte'] = num

        return query_dict

    def _query_tags(self):
        """ Find all structures matching given tags. """
        query_dict = dict()
        query_dict['$and'] = []
        for tag in self.args.get('tags'):
            temp_dict = dict()
            temp_dict['tags'] = dict()
            temp_dict['tags']['$in'] = [tag]
            query_dict['$and'].append(temp_dict)

        return query_dict

    def _query_doi(self):
        """ Find all structures matching given DOI,
        in format xxxx/xxxx.
        """
        doi = self.args.get('doi')
        if not isinstance(doi, list):
            doi = [doi]
        query_dict = dict()
        query_dict['doi'] = dict()
        query_dict['doi']['$in'] = doi

        return query_dict

    def _query_id(self):
        """ Find all structures matching given tags. """
        if isinstance(self.args.get('id'), str):
            self.args['id'] = self.args['id'].strip().split(' ')
        query_dict = dict()
        query_dict['text_id'] = self.args.get('id')
        return query_dict

    def _query_icsd(self):
        """ Find all structures matching given ICSD CollCode. """
        if not isinstance(self.args.get('icsd'), list):
            icsd = [self.args.get('icsd')]
        else:
            icsd = self.args.get('icsd')
        query_dict = dict()
        if isinstance(icsd[0], bool):
            query_dict['icsd'] = dict()
            query_dict['icsd']['$exists'] = icsd[0]
        elif icsd[0] == 0:
            query_dict['icsd'] = dict()
            query_dict['icsd']['$exists'] = True
        else:
            query_dict['$or'] = [{'icsd': {'$eq': str(icsd[0])}}, {'icsd': {'$eq': icsd[0]}}]
        return query_dict

    def _query_source(self):
        """ Find all structures with source string from args. """
        import re
        src_str = self.args.get('src_str')
        if not isinstance(src_str, list):
            src_str = [src_str]
        query_dict = dict()
        query_dict['source'] = dict()
        query_dict['source']['$in'] = [re.compile(src) for src in src_str]
        return query_dict

    def _query_root_source(self):
        """ Find all structures with root source string from args. """
        root_src = self.args.get('root_src')
        if not isinstance(root_src, list):
            root_src = [root_src]
        query_dict = dict()
        for src in root_src:
            query_dict['$or'] = []
            query_dict['$or'].append(dict())
            query_dict['$or'][-1]['root_source'] = src
        return query_dict

    @staticmethod
    def _query_available_values(field, cursor):
        """ Query the values stored under a particular field and
        print the information.

        Parameters:
            field (str): the field to query.
            cursor (list): the cursor to query.

        """
        supported_fields = [
            'doi',
            'tags',
            'root_source',
            'cnt_vector',
            'castep_version',
            'cut_off_energy'
        ]
        number_containing_field = sum([1 for doc in cursor if field in doc])
        if field in supported_fields and number_containing_field != 0:
            value_degeneracy = dict()
            for doc in cursor:
                if doc.get(field) is not None:
                    values = doc.get(field)
                    if isinstance(values, list):
                        for value in values:
                            if value in value_degeneracy:
                                value_degeneracy[value] += 1
                            else:
                                value_degeneracy[value] = 1

                    else:
                        if values in value_degeneracy:
                            value_degeneracy[values] += 1
                        else:
                            value_degeneracy[values] = 1

            print('Set of values under key {}:'.format(field))
            for value in sorted(value_degeneracy, key=value_degeneracy.get):
                print('{:<10} -> {:<10}'.format(value_degeneracy[value], value))
        else:
            print('Field {} unsupported for finding all possible values, must be one of {}'
                  .format(field, supported_fields))
            print('{}/{} contain field {}'.format(number_containing_field, len(cursor), field))

    @staticmethod
    def _query_quality():
        """ Find all structures with non-zero or non-existent (e.g.
        OQMD) quality.

        """
        query_dict = dict()
        query_dict['$or'] = []
        query_dict['$or'].append(dict())
        query_dict['$or'][-1]['quality'] = dict()
        query_dict['$or'][-1]['quality']['$gt'] = 0
        query_dict['$or'].append(dict())
        query_dict['$or'][-1]['quality'] = dict()
        query_dict['$or'][-1]['quality']['$exists'] = False

        return query_dict

    @staticmethod
    def _query_encap():
        """ Query only CNT encapsulated structures. """
        query_dict = dict()
        query_dict['encapsulated'] = dict()
        query_dict['encapsulated']['$exists'] = True

        return query_dict

    def _query_cnt_vector(self):
        """ Query structures within a nanotube of given chiral vector. """
        query_dict = dict()
        if not isinstance(self.args.get('cnt_vector'), list) or len(self.args.get('cnt_vector')) != 2:
            raise SystemExit('CNT vector query needs to be of form [n, m]')

        chiral_vec = self.args.get('cnt_vector')
        query_dict['cnt_chiral'] = dict()
        query_dict['cnt_chiral']['$eq'] = chiral_vec

        return query_dict

    def _query_sedc(self):
        """ Query all calculations using given SEDC scheme.

        Use --sedc null to query for no dispersion correction.

        """
        query_dict = dict()
        if self.args.get('sedc') != 'null':
            query_dict['sedc_scheme'] = self.args.get('sedc')
        else:
            query_dict['sedc_scheme'] = dict()
            query_dict['sedc_scheme']['$exists'] = False

        return query_dict

    def _query_xc_functional(self, xc_functional=None):
        """ Query all calculations with specified xc-functional.

        Keyword arguments:

            xc_functional (str): CASTEP string for xc-functional to
                override CLI.

        """
        query_dict = dict()
        if xc_functional is None:
            if isinstance(self.args.get('xc_functional'), list):
                xc_functional = self.args.get('xc_functional')[0]
            else:
                xc_functional = self.args.get('xc_functional')
        if xc_functional is not None:
            query_dict['xc_functional'] = xc_functional
        return query_dict

    def _query_spin(self):
        """ Query all calculations with spin polarisation,
        i.e. --spin n!=0, or non-spin-polarization, i.e. --spin 0.

        """
        query_dict = dict()
        if isinstance(self.args.get('spin'), list):
            spin = self.args.get('spin')[0]
        else:
            spin = self.args.get('spin')
        if spin == 'any':
            query_dict = dict()
        elif int(spin) == 0:
            query_dict['spin_polarized'] = dict()
            query_dict['spin_polarized']['$ne'] = True
        elif int(spin) > 0:
            query_dict['spin_polarized'] = True
        return query_dict

    def _query_calc(self, doc):
        """ Find all structures with matching
        accuracy to specified structure.
        """
        self._gs_enthalpy = doc['enthalpy_per_atom']

        query_dict = {}
        query_dict['$and'] = []
        query_dict['$and'].append(self._query_xc_functional(xc_functional=doc.get('xc_functional')))
        query_dict['$and'].append(self._query_float_range(
            'pressure', doc.get('pressure', 0.0), tolerance=self.args.get('pressure_tolerance') or 0.5))

        if self.args.get('time') is not None:
            query_dict['$and'].append(self._query_time())
        if 'spin_polarized' in doc and doc['spin_polarized']:
            if self.args.get('spin') != 'any':
                temp_dict = dict()
                temp_dict['spin_polarized'] = doc['spin_polarized']
                query_dict['$and'].append(temp_dict)
        else:
            if self.args.get('spin') != 'any':
                temp_dict = dict()
                temp_dict['spin_polarized'] = dict()
                temp_dict['spin_polarized']['$ne'] = True
                query_dict['$and'].append(temp_dict)
        if doc.get('grid_scale', 1.75) == 1.75:
            temp_dict = dict()
            temp_dict['$or'] = []
            temp_dict['$or'].append({'grid_scale': {'$exists': False}})
            temp_dict['$or'].append({'grid_scale': {'$eq': 1.75}})
            query_dict['$and'].append(temp_dict)
        else:
            temp_dict = dict()
            temp_dict['$or'] = []
            temp_dict['$or'].append({'grid_scale': {'$eq': doc.get('grid_scale')}})
            query_dict['$and'].append(temp_dict)
        if doc.get('fine_grid_scale', 1.75) == 1.75:
            temp_dict = dict()
            temp_dict['$or'] = []
            temp_dict['$or'].append({'fine_grid_scale': {'$exists': False}})
            temp_dict['$or'].append({'fine_grid_scale': {'$eq': 1.75}})
            query_dict['$and'].append(temp_dict)
        else:
            temp_dict = dict()
            temp_dict['$or'] = []
            temp_dict['$or'].append({'fine_grid_scale': {'$eq': doc.get('fine_grid_scale')}})
            query_dict['$and'].append(temp_dict)
        if 'geom_force_tol' in doc and doc['geom_force_tol'] != 0.05:
            temp_dict = dict()
            temp_dict['geom_force_tol'] = doc['geom_force_tol']
            query_dict['$and'].append(temp_dict)
        else:
            temp_dict = dict()
            temp_dict['$or'] = dict()
            temp_dict['$or'] = []
            temp_dict['$or'].append({'geom_force_tol': {'$exists': False}})
            temp_dict['$or'].append({'geom_force_tol': {'$eq': 0.05}})
            query_dict['$and'].append(temp_dict)
        if 'sedc_scheme' in doc:
            temp_dict = dict()
            temp_dict['sedc_scheme'] = doc['sedc_scheme']
            query_dict['$and'].append(temp_dict)
        else:
            temp_dict = dict()
            temp_dict['sedc_scheme'] = dict()
            temp_dict['sedc_scheme']['$exists'] = False
            query_dict['$and'].append(temp_dict)

        db = self.args.get('db')
        if isinstance(db, list):
            db = db[0]

        if self.args.get('loose') or (db is not None and 'oqmd' in db):
            return query_dict

        query_dict['$and'].append(self._query_float_range(
            'kpoints_mp_spacing', doc.get('kpoints_mp_spacing'), tolerance=self.args.get('kpoint_tolerance') or 0.01))
        query_dict['$and'].append(dict())
        query_dict['$and'][-1]['cut_off_energy'] = doc['cut_off_energy']

        if 'species_pot' in doc:
            for species in doc['species_pot']:
                temp_dict = dict()
                temp_dict['$or'] = []
                temp_dict['$or'].append(dict())
                temp_dict['$or'][-1]['species_pot.' + species] = dict()
                temp_dict['$or'][-1]['species_pot.' + species]['$exists'] = False
                temp_dict['$or'].append(dict())
                temp_dict['$or'][-1]['species_pot.' + species] = doc['species_pot'][species]
                query_dict['$and'].append(temp_dict)

        if self.debug:
            print('Calc match dict:')
            print(dumps(query_dict, indent=2))

        return query_dict

    def _query_time(self, since=False):
        """ Only include structures added before or after (depending on
        since) the date given in args['time'].

        Keyword arguments:
            since (bool): query before or after this time.

        """
        from datetime import datetime, timedelta
        from time import mktime
        query_dict = dict()
        time_period = timedelta(days=int(self.args.get('time')))
        time = (datetime.today() - time_period).timetuple()
        elapsed = str(hex(int(mktime(time))))[2:]
        cutoff_id = ObjectId(elapsed + '0000000000000000')
        query_dict['_id'] = dict()
        if since:
            query_dict['_id']['$gte'] = cutoff_id
        else:
            query_dict['_id']['$lte'] = cutoff_id

        return query_dict


class EmptyCursor:
    """ Empty cursor class for failures. """

    @staticmethod
    def count_documents(*args, **kwargs):
        """ Dummy function always returns 0. """
        return 0



from collections import defaultdict


class Settings:
    """ Light wrapper for global matador settings. """
    def __init__(self, settings=None):
        self.settings = defaultdict(dict)
        self.set = False

        if settings is not None:
            for key in settings:
                self.settings[key] = settings[key]
            self.set = True

    def __getitem__(self, key):
        return self.settings[key]

    def get(self, key, default=None):
        return self.settings.get(key, default)

    def __setitem__(self, key, val):
        if isinstance(val, dict):
            self.settings[key].update(val)
        else:
            self.settings[key] = val

    def __repr__(self):
        return str(self.settings)

    def reset(self):
        self.settings = defaultdict(dict)
        self.set = False



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements some light wrappers to
the Atomic Simulation Environment (ASE).

TODO: Use Crystal class to automate filling.

"""
import copy
import matador.crystal
from matador.utils.chem_utils import get_stoich
from matador.utils.cell_utils import get_spacegroup_spg

__all__ = ['ase2dict', 'doc2ase']


def ase2dict(atoms):
    """ Return a simple matador-style dictionary from
    an ase.Atoms object.

    Parameters:
        atoms (ase.Atoms): input structure.

    Returns:
        dict: matador output.

    """
    from matador.utils.cell_utils import cart2abc
    doc = {}

    # sort atoms, then their positions
    doc['atom_types'] = atoms.get_chemical_symbols()
    inds = [i[0] for i in sorted(enumerate(doc['atom_types']), key=lambda x: x[1])]
    doc['positions_frac'] = atoms.get_scaled_positions().tolist()
    doc['positions_frac'] = [doc['positions_frac'][ind] for ind in inds]
    doc['atom_types'] = [doc['atom_types'][ind] for ind in inds]
    try:
        doc['lattice_cart'] = atoms.get_cell().tolist()
    except AttributeError:
        doc['lattice_cart'] = atoms.get_cell().array.tolist()
    doc['lattice_abc'] = cart2abc(doc['lattice_cart'])
    doc['num_atoms'] = len(doc['atom_types'])
    doc['stoichiometry'] = get_stoich(doc['atom_types'])
    doc['cell_volume'] = atoms.get_volume()
    doc['elems'] = {atom for atom in doc['atom_types']}
    doc['num_fu'] = doc['num_atoms'] / int(sum(doc['stoichiometry'][i][1] for i in range(len(doc['stoichiometry']))))
    doc['space_group'] = get_spacegroup_spg(doc, symprec=0.001)

    return doc


def doc2ase(doc, add_keys_to_info=True):
    """ Convert matador document to simple ASE object.

    Parameters:
        doc (dict/:obj:`matador.crystal.Crystal`): matador document  or
            `Crystal` containing the structure.

    Keyword arguments:
        add_keys_to_info (bool): whether or not to add the keys from the
            matador document to the info section of the Atoms object.

    """
    from ase import Atoms

    atoms = Atoms(symbols=doc['atom_types'],
                  scaled_positions=doc['positions_frac'],
                  cell=doc['lattice_cart'],
                  pbc=True)

    if add_keys_to_info:
        if isinstance(doc, matador.crystal.Crystal):
            atoms.info['matador'] = doc._data
        else:
            atoms.info['matador'] = copy.deepcopy(doc)
        if '_id' in atoms.info['matador']:
            atoms.info['matador']['_id'] = str(atoms.info['matador']['_id'])

    return atoms



# coding: utf-8
# Distributed under the terms of the MIT License.

__all__ = ["GSAS_ATOMIC_SCATTERING_COEFFS", "RASPA_ATOMIC_SCATTERING_COEFFS"]

from .atomic_scattering.atomic_scattering_factors_gsas import GSAS_ATOMIC_SCATTERING_COEFFS
from .atomic_scattering.atomic_scattering_factors_raspa import RASPA_ATOMIC_SCATTERING_COEFFS



""" This file implements turning matador Crystal objects
into CrystalGraph objects.
"""
import networkx as nx
import numpy as np
import itertools

EPS = 1e-12


class CrystalGraph(nx.MultiDiGraph):
    def __init__(self, structure=None, graph=None, coordination_cutoff=1.1, bond_tolerance=1e20, num_images=1,
                 debug=False, separate_images=False, delete_one_way_bonds=False, max_bond_length=5):
        """ Create networkx.MultiDiGraph object with extra functionality for atomic networks.

        Keyword Arguments:
            structure (matador.Crystal):  crystal structure to network-ify
            graph (nx.MultiDiGraph): initialise from graph
            coordination_cutoff (float) : max multiplier of first
                coordination sphere for edge drawing
            num_images (int): number of periodic images to include in
                each direction
            separate_images (bool): whether or not to include image
                atoms as new nodes

        """

        super().__init__()

        if graph is None and structure is None:
            raise RuntimeError('No structure or graph to initialise network from.')

        if structure is not None:
            atoms = structure.sites
            num_atoms = len(atoms)
            element_bonds = {}
            images = list(itertools.product(range(-num_images, num_images+1), repeat=3))
            image_number = 0

            # now loop over pairs of atoms and decide whether to draw an edge
            from matador.utils.cell_utils import calc_pairwise_distances_pbc
            distances = calc_pairwise_distances_pbc(structure.positions_abs,
                                                    images,
                                                    structure.lattice_cart,
                                                    max_bond_length,
                                                    compress=False,
                                                    debug=True)

            # first over loop all pairs to find the minimum distance between all species pairs
            # and the minimum distance for each atom
            for i, atom in enumerate(atoms):
                self.add_node(i, species=atom.species)

            min_dists = [1e20 for atom in atoms]
            for index in np.where(~distances.mask)[0]:
                image_index = int(index / num_atoms**2)
                i = int((index - image_index * num_atoms**2) / num_atoms)
                j = int((index - image_index * num_atoms**2) % num_atoms)
                atom = atoms[i]
                other_atom = atoms[j]
                if (i == j and np.linalg.norm(images[image_index]) <= EPS):
                    continue
                dist = distances[index]
                pair_key = tuple(sorted([atom.species, other_atom.species]))
                if pair_key not in element_bonds or element_bonds[pair_key] > dist:
                    element_bonds[pair_key] = dist
                # find the closest image of an atom to i
                if dist < min_dists[i]:
                    min_dists[i] = dist

            if debug:
                print(min_dists)
                print(element_bonds)

            for index in np.where(~distances.mask)[0]:
                image_index = int(index / num_atoms**2)
                i = int((index - image_index * num_atoms**2) / num_atoms)
                j = int((index - image_index * num_atoms**2) % num_atoms)
                atom = atoms[i]
                other_atom = atoms[j]
                min_dist = min_dists[i]
                if (i == j and np.linalg.norm(images[image_index]) <= EPS):
                    continue
                dist = distances[index]
                pair_key = tuple(sorted([atom.species, other_atom.species]))
                if dist <= min_dist*coordination_cutoff and dist <= element_bonds[pair_key]*bond_tolerance:
                    if separate_images and all([val <= 0+1e-8 for val in images[image_index]]):
                        image_number += 1
                        self.add_node(j+image_number, species=atoms[j].species)
                        self.add_edge(i, j+image_number, dist=dist)
                        self.add_edge(j+image_number, i, dist=dist)
                    else:
                        is_image = np.linalg.norm(images[image_index]) > EPS
                        self.add_edge(i, j, dist=dist, image=is_image)

        elif graph is not None:
            for node, data in graph.nodes.data():
                self.add_node(node, species=data['species'], image=data.get('image', False))
            for node_in, node_out, data in graph.edges.data():
                if not delete_one_way_bonds or (node_out, node_in) in graph.edges():
                    self.add_edge(node_in, node_out, dist=data.get('dist', 0), image=data.get('image', False))

    def get_strongly_connected_component_subgraphs(self, delete_one_way_bonds=True):
        """ Return generator of strongly-connected subgraphs in CrystalGraph format. """
        return (CrystalGraph(graph=self.subgraph(c),
                             delete_one_way_bonds=delete_one_way_bonds)
                for c in nx.strongly_connected_components(self))

    def get_communities(self, graph=None, **louvain_kwargs):
        """ Return list of community subgraphs in CrystalGraph format. """
        import community as louvain

        if graph is None:
            graph = self

        if graph.is_directed():
            undirected_graph = self.remove_directionality(graph=graph)

        partition = louvain.best_partition(undirected_graph, **louvain_kwargs)
        size = len(set(partition.values()))
        subgraphs = [nx.MultiGraph() for i in range(size)]

        for node in partition:
            subgraphs[partition[node]].add_node(
                node, species=list(self.nodes(data=True))[list(self.nodes()).index(node)][1]['species']
            )
        for edge in self.edges():
            if partition[edge[0]] == partition[edge[1]]:
                subgraphs[partition[edge[0]]].add_edge(edge[0], edge[1])

        subgraphs = [CrystalGraph(graph=sg) for sg in subgraphs]

        return subgraphs, partition

    def remove_directionality(self, graph=None):
        if graph is None:
            graph = self
        import networkx as nx
        undirected_graph = nx.MultiGraph()
        for node in graph.nodes(data=True):
            undirected_graph.add_node(node[0], species=node[1]['species'])
        for edge in graph.edges(data=True):
            if (edge[1], edge[0]) not in undirected_graph.edges():
                undirected_graph.add_edge(edge[0], edge[1], dist=edge[2].get('dist', 0), image=edge[2].get('image', False))

        return undirected_graph

    def set_unique_subgraphs(self, method='community'):
        """ Filter strongly connected component subgraphs for isomorphism with others inside
        CrystalGraph. Sets self.unique_subgraph to a set of such subgraphs.
        """
        if method == 'community':
            self.unique_subgraphs = get_unique_subgraphs(self.get_communities())
        elif method == 'strongly_connected':
            self.unique_subgraphs = get_unique_subgraphs(self.get_strongly_connected_component_subgraphs())
        elif method == 'both':
            strong_subgraphs = self.get_strongly_connected_component_subgraphs()
            community_subgraphs = []
            for sg in strong_subgraphs:
                community_subgraphs.extend(sg.get_communities())

            self.unique_subgraphs = get_unique_subgraphs(community_subgraphs)

    def get_bonds_per_atom(self):
        num_bonds = 0
        for node_in in self.nodes():
            for node_out in self.nodes():
                if node_in == node_out:
                    continue
                if self.has_edge(node_in, node_out) and self.has_edge(node_out, node_in):
                    num_bonds += 1
        return num_bonds / self.number_of_nodes()


def node_match(n1, n2):
    return n1['species'] == n2['species']


def get_unique_subgraphs(subgraphs):
    """ Filter strongly connected component subgraphs for isomorphism with others.

    Input:

        | subgraphs: list(CrystalGraph), list of subgraph objects to filter

    Returns:

        | unique_subgraphs: set(CrystalGraph), set of unique subgraphs

    """
    unique_subgraphs = set()
    for subgraph in subgraphs:
        if not any([are_graphs_the_same(subgraph, other_subgraph) for other_subgraph in unique_subgraphs]):
            unique_subgraphs.add(subgraph)
    return unique_subgraphs


def are_graphs_the_same(g1, g2, edge_match=None):
    if edge_match is None:
        def edge_match(e1, e2):
            atol = 0.1
            rtol = 0.05
            return abs(e1[0]['dist'] - e2[0]['dist']) <= atol + rtol * e2[0]['dist'] and e1[0]['image'] == e2[0]['image']

    return nx.is_isomorphic(g1, g2,
                            node_match=lambda n1, n2: n1['species'] == n2['species'],
                            edge_match=edge_match)


def draw_network(structure,
                 layout=None, edge_labels=False, node_index=False,
                 curved_edges=True, node_colour='elem', partition=None, ax=None):
    import networkx as nx
    from matador.utils.viz_utils import get_element_colours
    import matplotlib.pyplot as plt

    element_colours = get_element_colours()
    try:
        network = structure.network
    except Exception:
        network = structure
    if layout is None:
        pos = nx.spring_layout(network)
    else:
        pos = layout

    if ax is None:
        fig, ax = plt.subplots()

    if node_colour == 'degree':
        coords = list(set(dict(network.degree).values()))
        cmap = plt.cm.get_cmap('Dark2', len(coords)).colors
        colours = [cmap[coords.index(network.degree[node])] for node in network.nodes()]

    elif node_colour == 'partition' and partition is not None:
        num_partitions = len(set(partition.values()))
        cmap = plt.cm.get_cmap('Dark2', num_partitions).colors
        colours = [cmap[partition[node]] for node in network.nodes()]

    else:
        elem_map = element_colours
        colours = [elem_map.get(data['species']) for node, data in network.nodes.data()]

    if node_index:
        labels = {node: '{} \\#{}'.format(data['species'], node) for node, data in network.nodes.data()}
    else:
        labels = {node: str(data['species']) for node, data in network.nodes.data()}

    edge_colours = []
    for edge in network.edges(data=True):
        if edge[2].get('image', True):
            edge_colours.append('grey')
        else:
            edge_colours.append('black')

    nx.draw_networkx_nodes(network, pos, node_color=colours, edgecolors='black', linewidths=2, node_size=1000, ax=ax)
    nx.draw_networkx_edges(network, pos, edge_color=edge_colours, linewidths=2, node_size=1000, ax=ax)
    if edge_labels:
        edge_weight = dict()
        for edge in network.edges(data=True):
            # data = edge[2]
            edge = (edge[0], edge[1])
            if edge not in edge_weight and (edge[1], edge[0]) not in edge_weight:
                edge_weight[edge] = 1
            else:
                if edge in edge_weight:
                    edge_weight[edge] += 1
                else:
                    edge_weight[(edge[1], edge[0])] += 1
        edge_label_dict = edge_weight
        nx.draw_networkx_edge_labels(network, pos, edge_labels=edge_label_dict, ax=ax)
    nx.draw_networkx_labels(network, pos, labels=labels, ax=ax)
    plt.axis('off')



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The query module provides the DBQuery class that implements all
database queries, either for individual compositions/formulae, or for
matching calculations at the same accuracy to create phase diagrams.

"""


__all__ = ['DBQuery']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.query.query import DBQuery
